# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Error reporting.
"""

from __future__ import annotations

import dataclasses
import enum
import functools
import os
import sys
import types
import typing as t

import nox

from ...messages import Level, Location, Message
from . import IN_CI, nox_has_color

if t.TYPE_CHECKING:
    Formatter = t.Callable[[list[Message]], t.Generator[str]]


OutputFormat = t.Literal["plain", "fancy"]


@dataclasses.dataclass(frozen=True, eq=False, repr=False)
class Markings:
    """
    Collection of marking symbols.
    """

    line_indicator: str
    pos_indicator: str
    underline_indicator: str
    box_left_top: str
    box_left: str
    box_left_bottom: str
    box_bottom: str
    box_right_bottom: str
    box_right: str
    box_right_top: str
    box_top: str
    horizontal_ellipsis: str
    vertical_ellipsis: str


ASCII_MARKINGS = Markings(
    line_indicator=">",
    pos_indicator="^",
    underline_indicator="^",
    box_left_top="/",
    box_left="|",
    box_left_bottom="\\",
    box_bottom="-",
    box_right_bottom="/",
    box_right="|",
    box_right_top="\\",
    box_top="-",
    horizontal_ellipsis="...",
    vertical_ellipsis=":",
)

UNICODE_MARKINGS = Markings(
    # https://en.wikipedia.org/wiki/Arrows_%28Unicode_block%29
    line_indicator="→",
    pos_indicator="↑",
    underline_indicator="▔",
    # https://en.wikipedia.org/wiki/Box-drawing_characters
    box_left_top="┌",
    box_left="│",
    box_left_bottom="└",
    box_bottom="─",
    box_right_bottom="┘",
    box_right="│",
    box_right_top="┐",
    box_top="─",
    horizontal_ellipsis="…",
    vertical_ellipsis="⋮",
)


@functools.cache
def get_box_markings() -> Markings:
    """
    Determine how to draw boxes.
    """
    if os.getenv("ANTSIBULL_BOX_MARKINGS") == "ascii":
        return ASCII_MARKINGS
    return UNICODE_MARKINGS


@functools.cache
def get_output_format() -> OutputFormat:
    """
    Determine the output format.
    """
    output_format = os.getenv("ANTSIBULL_OUTPUT_FORMAT")
    if output_format in t.get_args(OutputFormat):
        return t.cast(OutputFormat, output_format)
    if IN_CI:
        return "plain"
    return "fancy"


def split_lines(text: str) -> list[str]:
    """
    Given a text with newlines, split it up into single lines.
    """
    return text.rstrip("\n").splitlines()


def split_lines_with_prefix(
    text: str,
    *,
    prefix: str = "",
    separator: str = " ",
    at_least_one_line: bool = False,
) -> t.Generator[str]:
    """
    Given a text with newlines, emit single lines with optional prefix.

    By default prefix and line are separated by a single space.
    This can be changed by passing an appropriate ``separator``.
    """
    lines = split_lines(text)
    for index, line in enumerate(lines):
        if index == 1:
            prefix = " " * len(prefix)
        yield f"{prefix}{separator}{line}"
    if not lines and at_least_one_line:
        yield prefix


class SynchronizedOutput:
    """
    Print output to stdout, but keep it synchronized to stderr.
    """

    def __init__(self) -> None:
        self._has_output = False

    @property
    def has_output(self) -> bool:
        """
        Whether any output has been emitted.
        """
        return self._has_output

    def msg(self, message: str) -> None:
        """
        Print a one-line message.
        """
        if not self._has_output:
            sys.stderr.flush()
            self._has_output = True
        print(message)

    def __enter__(self) -> t.Self:
        return self

    def __exit__(
        self,
        exc_type: t.Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> t.Literal[False]:
        if self._has_output:
            sys.stdout.flush()
        return False


def should_fail(messages: list[Message]) -> bool:
    """
    Determine whether a test with the given list of output messages should fail.
    """
    return any(message.level in (Level.WARNING, Level.ERROR) for message in messages)


def format_messages_plain(messages: list[Message]) -> t.Generator[str]:
    """
    Format a list of messages as a sequence of lines.
    """
    for message in sorted(messages):
        loc_line = "0"
        loc_column = 0
        if message.position is not None:
            loc_line = str(message.position.line)
            loc_column = message.position.column or 0
            if not message.position.exact:
                loc_line = f"~{loc_line}"
        prefix = f"{message.file or ''}:{loc_line}:{loc_column}:"
        if message.id is not None:
            prefix = f"{prefix} [{message.id}]"
        content = message.message
        if message.symbol is not None:
            content = f"{content} [{message.symbol}]"
        if message.hint is not None:
            content = f"{content}\n{message.hint}"
        if message.note is not None:
            content = f"{content}\nNote: {message.note}"
        yield from split_lines_with_prefix(
            content, prefix=prefix, at_least_one_line=True
        )


_RESET = "\x1b[0m"
_BOLD = "\x1b[1m"
_FAINT = "\x1b[2m"
_ITALICS = "\x1b[3m"
_UNDERLINE = "\x1b[4m"


class Color(enum.Enum):
    """
    ANSI colors.
    """

    NONE = ""
    BLACK = "\x1b[30m"
    RED = "\x1b[31m"
    GREEN = "\x1b[32m"
    YELLOW = "\x1b[33m"
    BLUE = "\x1b[34m"
    MAGENTA = "\x1b[35m"
    CYAN = "\x1b[36m"
    WHITE = "\x1b[37m"
    BRIGHT_BLACK = "\x1b[90m"
    BRIGHT_RED = "\x1b[91m"
    BRIGHT_GREEN = "\x1b[92m"
    BRIGHT_YELLOW = "\x1b[93m"
    BRIGHT_BLUE = "\x1b[94m"
    BRIGHT_MAGENTA = "\x1b[95m"
    BRIGHT_CYAN = "\x1b[96m"
    BRIGHT_WHITE = "\x1b[97m"


@dataclasses.dataclass(frozen=True)
class _Formatting:
    color: Color = Color.NONE
    bold: bool = False
    faint: bool = False
    italics: bool = False
    underline: bool = False

    def update(
        self,
        *,
        color: Color | None = None,
        bold: bool | None = None,
        faint: bool | None = None,
        italics: bool | None = None,
        underline: bool | None = None,
    ) -> _Formatting:
        """
        Update attributes.
        """
        if bold and faint:
            raise ValueError("Text cannot be both bold and faint")
        new_bold = self.bold
        new_faint = self.faint
        if bold is not None:
            new_bold = bold
            if new_bold:
                new_faint = False
        if faint is not None:
            new_faint = faint
            if new_faint:
                new_bold = False
        return _Formatting(
            self.color if color is None else color,
            new_bold,
            new_faint,
            self.italics if italics is None else italics,
            self.underline if underline is None else underline,
        )

    def eq_color(self, other: _Formatting) -> bool:
        """Compare color attributes of two formattings."""
        return self.color == other.color

    def eq_formatting(self, other: _Formatting) -> bool:
        """Compare formatting attributes of two formattings."""
        return (
            self.bold == other.bold
            and self.faint == other.faint
            and self.italics == other.italics
            and self.underline == other.underline
        )

    def _new_format(
        self,
        *,
        use_color: bool,
        use_formatting: bool,
    ) -> list[str]:
        result = []
        if use_formatting:
            if self.bold:
                result.append(_BOLD)
            if self.faint:
                result.append(_FAINT)
            if self.italics:
                result.append(_ITALICS)
            if self.underline:
                result.append(_UNDERLINE)
        if use_color and self.color is not Color.NONE:
            result.append(self.color.value)
        return result

    def _update_format(
        self,
        *,
        use_color: bool,
        use_formatting: bool,
        previous: _Formatting,
    ) -> list[str]:
        # Determine whether need a reset
        needs_reset = False
        if use_formatting and not needs_reset:
            needs_reset = (
                (not self.bold and previous.bold)
                or (not self.faint and previous.faint)
                or (not self.italics and previous.italics)
                or (not self.underline and previous.underline)
            )
        if use_color and not needs_reset:
            needs_reset = self.color != previous.color and self.color is Color.NONE

        # Compose codes
        result = []
        if needs_reset:
            result.append(_RESET)
        if use_formatting:
            if self.bold and (needs_reset or self.bold != previous.bold):
                result.append(_BOLD)
            if self.faint and (needs_reset or self.faint != previous.faint):
                result.append(_FAINT)
            if self.italics and (needs_reset or self.italics != previous.italics):
                result.append(_ITALICS)
            if self.underline and (needs_reset or self.underline != previous.underline):
                result.append(_UNDERLINE)
        if (
            use_color
            and self.color is not Color.NONE
            and (needs_reset or self.color != previous.color)
        ):
            result.append(self.color.value)
        return result

    def format(
        self,
        *,
        use_color: bool,
        use_formatting: bool,
        previous: _Formatting | None = None,
    ) -> list[str]:
        """
        Compute format update as sequence of ANSI codes.
        """
        if previous is None:
            return self._new_format(use_color=use_color, use_formatting=use_formatting)
        return self._update_format(
            use_color=use_color, use_formatting=use_formatting, previous=previous
        )


_CLEAR = _Formatting()


class ColorComposer:
    """
    Compose text that uses optional colors and formatting.
    """

    def __init__(self, *, use_color: bool, use_formatting: bool) -> None:
        """
        Set up color and formatting composer.
        """
        self.use_color = use_color
        self.use_formatting = use_formatting
        self.last_formatting = _CLEAR
        self.formatting = _CLEAR
        self.output: list[str] = []

    def format(
        self,
        *,
        color: Color | None = None,
        bold: bool | None = None,
        faint: bool | None = None,
        italics: bool | None = None,
        underline: bool | None = None,
    ) -> t.Self:
        """
        Adjust formatting.
        """
        self.formatting = self.formatting.update(
            color=color, bold=bold, faint=faint, italics=italics, underline=underline
        )
        return self

    def add_text(
        self,
        text: str,
        *,
        color: Color | None = None,
        bold: bool | None = None,
        faint: bool | None = None,
        italics: bool | None = None,
        underline: bool | None = None,
    ) -> t.Self:
        """
        Add text with optional formatting adjustment.
        """
        self.format(
            color=color, bold=bold, faint=faint, italics=italics, underline=underline
        )
        if text:
            self.output.extend(
                self.formatting.format(
                    use_color=self.use_color,
                    use_formatting=self.use_formatting,
                    previous=self.last_formatting,
                )
            )
            self.last_formatting = self.formatting
            self.output.append(text)
        return self

    @property
    def empty(self) -> bool:
        """
        Whether the formatter already has output.
        """
        return not self.output

    @property
    def value(self) -> str:
        """
        Obtain formatted value.
        """
        result = "".join(self.output)
        if (self.use_color and not self.last_formatting.eq_color(_CLEAR)) or (
            self.use_formatting and not self.last_formatting.eq_formatting(_CLEAR)
        ):
            result = f"{result}{_RESET}"
        return result


class _ContentProvider:
    def __init__(self) -> None:
        self.filename: str | None = None
        self.content: list[str] | None = None

    def _ensure_loaded(self, filename: str) -> None:
        if self.filename == filename:
            return
        self.filename = filename
        self.content = None
        try:
            with open(filename, "rt", encoding="utf-8") as f:
                self.content = f.read().splitlines()
        except Exception:  # pylint: disable=broad-exception-caught
            # This can be caused by I/O errors, decoding errors, out of memory errors, ...
            # In all these cases, simply do not show the content.
            pass

    @staticmethod
    def _expand(line: str, *, tabulator_width: int = 8) -> str:
        result = []
        idx = 0
        for idx, part in enumerate(line.split("\t")):
            if idx:
                act_tab_w = tabulator_width - (idx % tabulator_width)
                if act_tab_w == 0:
                    act_tab_w = tabulator_width
                result.append(" " * act_tab_w)
                idx += act_tab_w
            result.append(part)
            idx += len(part)
        return "".join(result)

    @staticmethod
    def _get_column(
        lines: list[str],
        line: int,
        column: int | None,
        *,
        tabulator_width: int = 8,
        is_end: bool = False,
    ) -> int | None:
        if column is None:
            return None
        if not 0 <= line < len(lines):
            return None
        column -= 1
        idx = 0
        for idx, part in enumerate(lines[line].split("\t")):
            if idx:
                act_tab_w = tabulator_width - (idx % tabulator_width)
                if act_tab_w == 0:
                    act_tab_w = tabulator_width
                if column == 0:
                    return ((idx + act_tab_w - 1) if is_end else idx) + 1
                idx += act_tab_w
                column -= 1
            part_len = len(part)
            if column < part_len:
                return idx + column + 1
            idx += part_len
            column -= part_len
        return idx + column + 1

    @staticmethod
    def _find(index: int, direction: int, content: list[str]) -> int:
        length = len(content)
        candidate = index
        for _ in range(2):
            candidate += direction
            if candidate < 0 or candidate >= length:
                break
            if content[candidate].strip():
                return candidate
        candidate = index + direction
        if candidate < 0 or candidate >= length:
            return index
        return candidate

    def get_content_lines(
        self, *, filename: str, start: Location, end: Location | None
    ) -> (
        tuple[
            list[tuple[int | None, str]],
            tuple[int, int | None],
            tuple[int, int | None] | None,
        ]
        | None
    ):
        """
        Return lines containing the start and end locations.
        """
        self._ensure_loaded(filename)
        if self.content is None:
            return None
        last_idx = first_idx = start.line - 1
        if end is not None:
            first_idx = min(first_idx, end.line - 1)
            last_idx = max(last_idx, end.line - 1)
        act_first_idx = first_idx
        act_last_idx = last_idx
        first_idx = self._find(first_idx, -1, self.content)
        last_idx = self._find(last_idx, +1, self.content)
        result: list[tuple[int | None, str]] = []
        if first_idx > 0:
            result.append((None, ""))
        if act_last_idx - act_first_idx > 5:
            for idx in range(first_idx, act_first_idx + 2):
                result.append((idx + 1, self._expand(self.content[idx])))
            result.append((None, ""))
            for idx in range(act_last_idx - 1, last_idx + 1):
                result.append((idx + 1, self._expand(self.content[idx])))
        else:
            for idx in range(first_idx, last_idx + 1):
                result.append((idx + 1, self._expand(self.content[idx])))
        if last_idx + 1 < len(self.content):
            result.append((None, ""))
        return (
            result,
            (start.line, self._get_column(self.content, start.line - 1, start.column)),
            (
                None
                if end is None
                else (
                    end.line,
                    self._get_column(self.content, end.line - 1, end.column),
                )
            ),
        )


def _determine_marker(
    inside: bool,
    line_no: int | None,
    line_length: int,
    content_start: tuple[int, int | None],
    content_end: tuple[int, int | None] | None,
    extra_indent: str,
    markings: Markings,
) -> tuple[bool, bool, str, str, str]:
    last_inside = False
    before_mark = ""
    after_mark = ""
    this_extra_indent = extra_indent
    if line_no == content_start[0]:
        inside = True
        if content_end is None:
            last_inside = True
            if content_start[1] is None:
                this_extra_indent = f"{markings.line_indicator} "
            else:
                after_mark = (
                    extra_indent + " " * (content_start[1] - 1) + markings.pos_indicator
                )
        elif content_end[0] == content_start[0]:
            last_inside = True
            if content_start[1] is not None and content_end[1] is not None:
                end_col = content_end[1]
                if end_col <= 0:
                    end_col = max(content_start[1], line_length + 1)
                mark_len = end_col - content_start[1] + 1
                after_mark = (
                    extra_indent
                    + " " * (content_start[1] - 1)
                    + (
                        markings.pos_indicator
                        if mark_len == 1
                        else markings.underline_indicator * mark_len
                    )
                )
            else:
                before_mark = markings.box_left_top
                this_extra_indent = f"{markings.box_left} "
                after_mark = markings.box_left_bottom
        else:
            before_mark = markings.box_left_top
            this_extra_indent = f"{markings.box_left} "
    elif content_end is not None and line_no == content_end[0]:
        last_inside = True
        this_extra_indent = f"{markings.box_left} "
        after_mark = markings.box_left_bottom
        if content_end[1] is not None:
            end_col = content_end[1]
            if end_col <= 0:
                end_col = line_length + 1
            after_mark += (
                markings.box_bottom * (end_col + 1) + markings.box_right_bottom
            )
    elif inside:
        this_extra_indent = f"{markings.box_left} "
    return inside, last_inside, before_mark, after_mark, this_extra_indent


def _render_code(
    content: list[tuple[int | None, str]],
    content_start: tuple[int, int | None],
    content_end: tuple[int, int | None] | None,
    mkcomp: t.Callable[[], ColorComposer],
    markings: Markings,
) -> t.Generator[str]:
    assert content
    line_prefixes = [
        markings.vertical_ellipsis if line_no is None else str(line_no)
        for line_no, _ in content
    ]
    max_prefix_len = max(len(prefix) for prefix in line_prefixes)
    empty = " " * max_prefix_len
    line_prefixes = [
        " " * (max_prefix_len - len(prefix)) + prefix for prefix in line_prefixes
    ]
    extra_indent = ""
    if (
        content_end is not None and content_start[0] != content_end[0]
    ) or content_start[1] is None:
        extra_indent = "  "
    inside = False
    for prefix, (line_no, line) in zip(line_prefixes, content):
        comp = mkcomp()
        comp.add_text(prefix, faint=True)
        comp.add_text(f" {markings.box_left} ", faint=True)
        inside, last_inside, before_mark, after_mark, this_extra_indent = (
            _determine_marker(
                inside,
                line_no,
                len(line),
                content_start,
                content_end,
                extra_indent,
                markings,
            )
        )
        comp.add_text(
            this_extra_indent,
            color=None if this_extra_indent is extra_indent else Color.RED,
            faint=False,
        )
        comp.add_text(line, color=Color.NONE)
        if before_mark:
            before_comp = mkcomp()
            before_comp.add_text(empty, faint=True)
            before_comp.add_text(f" {markings.box_left} ", faint=True)
            before_comp.add_text(before_mark, faint=False, color=Color.RED)
            yield before_comp.value
        yield comp.value
        if last_inside:
            inside = False
        if after_mark:
            after_comp = mkcomp()
            after_comp.add_text(empty, faint=True)
            after_comp.add_text(f" {markings.box_left} ", faint=True)
            after_comp.add_text(after_mark, faint=False, color=Color.RED)
            yield after_comp.value


def _compose_first_line(
    message: Message, mkcomp: t.Callable[[], ColorComposer]
) -> ColorComposer:
    comp = mkcomp()
    if message.file:
        comp.add_text(message.file or "")
        if message.position is not None:
            comp.add_text(
                f":{'' if message.position.exact else '~'}{message.position.line}"
            )
            if message.position.column is not None:
                comp.add_text(f":{message.position.column}")
        comp.add_text(":")
    if message.id is not None:
        if not comp.empty:
            comp.add_text(" ")
        comp.add_text(f"[{message.id}]", color=Color.RED)
    return comp


def _compose_message_with_note(
    message: Message,
    indent: str,
    mkcomp: t.Callable[[], ColorComposer],
    *,
    add_hint: bool,
    at_least_one_line: bool,
) -> t.Generator[str]:
    msg_content = message.message
    if message.symbol is not None:
        msg_content = f"{msg_content} [{message.symbol}]"
    if message.hint is not None and add_hint:
        msg_content = f"{msg_content}\n{message.hint}"
    for line in split_lines(msg_content):
        comp = mkcomp()
        comp.add_text(indent)
        comp.add_text(line, bold=True)
        yield comp.value
        at_least_one_line = False
    if message.note is not None:
        for index, line in enumerate(split_lines(message.note)):
            comp = mkcomp()
            comp.add_text(indent)
            if index == 0:
                comp.add_text("Note: ", italics=True)
            comp.add_text(line, italics=False)
            yield comp.value
            at_least_one_line = False
    if at_least_one_line:
        yield indent


def format_messages_with_context(
    messages: list[Message], *, color: bool, markings: Markings
) -> t.Generator[str]:
    """
    Format a list of messages as a sequence of lines with context and optional color.
    """

    def mkcomp() -> ColorComposer:
        return ColorComposer(use_color=color, use_formatting=color)

    content_provider = _ContentProvider()
    for index, message in enumerate(sorted(messages)):
        if index:
            yield ""

        # First line: file, start position, ID
        comp = _compose_first_line(message, mkcomp)
        has_first_line = not comp.empty
        if has_first_line:
            yield comp.value

        content: list[tuple[int | None, str]] | None = None
        content_start: tuple[int, int | None] = (0, None)
        content_end: tuple[int, int | None] | None = None
        if message.file and message.position and message.position.exact:
            content_tuple = content_provider.get_content_lines(
                filename=message.file,
                start=message.position,
                end=(
                    message.end_position
                    if message.end_position and message.end_position.exact
                    else None
                ),
            )
            if content_tuple is not None:
                content, content_start, content_end = content_tuple

        # Afterwards: message with note, and hint if content isn't shown
        indent = "  " if has_first_line else ""
        yield from _compose_message_with_note(
            message,
            indent,
            mkcomp,
            add_hint=content is None,
            at_least_one_line=not has_first_line and not content,
        )

        if content is not None:
            yield from _render_code(
                content, content_start, content_end, mkcomp, markings
            )

            if message.hint is not None:
                for line in split_lines(message.hint):
                    comp = mkcomp()
                    comp.add_text(indent)
                    comp.add_text(line, color=Color.GREEN, bold=True)
                    yield comp.value


def get_formatter(
    session: nox.Session,
    *,
    output_format: OutputFormat | None = None,
    use_color: bool | None = None,
    markings: Markings | None = None,
) -> Formatter:
    """
    Return the suggested message formatter to use.
    """
    if output_format is None:
        output_format = get_output_format()
    if output_format == "plain":
        return format_messages_plain
    if output_format == "fancy":
        if markings is None:
            markings = get_box_markings()
        return lambda messages: format_messages_with_context(
            messages,
            color=nox_has_color(session) if use_color is None else use_color,
            markings=markings,
        )
    raise AssertionError(f"Invalid output format {output_format}")


def print_messages(
    *, session: nox.Session, messages: list[Message], fail_msg: str
) -> None:
    """
    Print messages, and error out if at least one error has been found.
    """
    with SynchronizedOutput() as output:
        for line in get_formatter(session)(messages):
            output.msg(line)
    if should_fail(messages):
        session.error(fail_msg)


__all__ = (
    "Markings",
    "ASCII_MARKINGS",
    "UNICODE_MARKINGS",
    "get_box_markings",
    "get_output_format",
    "split_lines_with_prefix",
    "SynchronizedOutput",
    "should_fail",
    "format_messages_plain",
    "format_messages_with_context",
    "get_formatter",
    "print_messages",
)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Utils for creating nox sessions.
"""

from __future__ import annotations

import argparse
import logging
import os
import sys
import typing as t
from contextlib import contextmanager

import nox
from nox.logger import OUTPUT as nox_OUTPUT

# https://docs.github.com/en/actions/writing-workflows/choosing-what-your-workflow-does/store-information-in-variables#default-environment-variables
# https://docs.gitlab.com/ci/variables/predefined_variables/#predefined-variables
# https://docs.travis-ci.com/user/environment-variables/#default-environment-variables
IN_CI = os.environ.get("CI") == "true"
IN_GITHUB_ACTIONS = bool(os.environ.get("GITHUB_ACTION"))

_SESSIONS: dict[str, list[dict[str, t.Any]]] = {}


def nox_has_verbosity() -> bool:
    """
    Determine whether nox is run with verbosity enabled.
    """
    logger = logging.getLogger()
    return logger.level <= nox_OUTPUT


@contextmanager
def silence_run_verbosity() -> t.Iterator[None]:
    """
    When using session.run() with silent=True, nox will log the output
    if -v is used. Using this context manager prevents printing the output.
    """
    logger = logging.getLogger()
    original_level = logger.level
    try:
        logger.setLevel(max(nox_OUTPUT + 1, original_level))
        yield
    finally:
        logger.setLevel(original_level)


def nox_has_color(session: nox.Session) -> bool:
    """
    Determine whether nox is run with color mode.
    """
    # I don't know of a better way to obtain this information.
    # It is also stored in the logging stream handler that nox
    # installs, but extracting it from there seems even more hacky...
    return session._runner.global_config.color  # pylint: disable=protected-access


@contextmanager
def ci_group(name: str) -> t.Iterator[tuple[str, bool]]:
    """
    Try to ensure that the output inside the context is printed in a collapsable group.

    This is highly CI system dependent, and currently only works for GitHub Actions.
    """
    sys.stderr.flush()
    is_collapsing = False
    if IN_GITHUB_ACTIONS:
        print(f"::group::{name}")
        sys.stdout.flush()
        is_collapsing = True
    yield ("  " if is_collapsing else "", is_collapsing)
    sys.stderr.flush()
    if IN_GITHUB_ACTIONS:
        print("::endgroup::")
    sys.stdout.flush()


def register(name: str, data: dict[str, t.Any]) -> None:
    """
    Register a session name for matrix generation with additional data.
    """
    if name not in _SESSIONS:
        _SESSIONS[name] = []
    _SESSIONS[name].append(data)


def get_registered_sessions() -> dict[str, list[dict[str, t.Any]]]:
    """
    Return all registered sessions.
    """
    return {
        name: [session.copy() for session in sessions]
        for name, sessions in _SESSIONS.items()
    }


def compose_description(
    *,
    prefix: str | dict[t.Literal["one", "other"], str] | None = None,
    programs: dict[str, str | bool | None],
) -> str:
    """
    Compose a description for a nox session from several configurable parts.
    """
    parts: list[str] = []

    def add(text: str, *, comma: bool = False) -> None:
        if parts:
            if comma:
                parts.append(", ")
            else:
                parts.append(" ")
        parts.append(text)

    active_programs = [
        (program, value if isinstance(value, str) else None)
        for program, value in programs.items()
        if value not in (False, None)
    ]

    if prefix:
        if isinstance(prefix, dict):
            if len(active_programs) == 1 and "one" in prefix:
                add(prefix["one"])
            else:
                add(prefix["other"])
        else:
            add(prefix)

    for index, (program, value) in enumerate(active_programs):
        if index + 1 == len(active_programs) and index > 0:
            add("and", comma=index > 1)
        add(program, comma=index > 0 and index + 1 < len(active_programs))
        if value is not None:
            add(f"({value})")

    return "".join(parts)


def parse_args(
    *,
    session: nox.Session,
    parser: argparse.ArgumentParser,
) -> argparse.Namespace | None:
    """
    Parse arguments for a session.

    The ``ArgumentParser`` object should be created with ``exit_on_error=False``.

    Return ``None`` if the session should exit immediately, or a ``argparse.Namespace``
    object with the parsed arguments.
    """
    try:
        return parser.parse_args(session.posargs)
    except argparse.ArgumentError as exc:
        # session.error() never returns, but pylint doesn't seem to know that,
        # so we use the 'return function()' pattern...
        return session.error(str(exc))
    except SystemExit as exc:
        if exc.code in (0, None):
            return None
        if isinstance(exc.code, str):
            session.error(exc.code)
        return session.error("Error")


def normalize_session_name(name: str) -> str:
    """
    Replace/remove not allowed characters in session names.
    """
    # So far, I'm only aware of '/'.
    return name.replace("/", "-")


__all__ = [
    "ci_group",
    "compose_description",
    "get_registered_sessions",
    "normalize_session_name",
    "nox_has_verbosity",
    "register",
    "silence_run_verbosity",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Interpret config.
"""

from __future__ import annotations

import typing as t

import nox

from .ansible import AnsibleCoreVersion
from .collection import CollectionSource, setup_collection_sources
from .config import ActionGroup as ConfigActionGroup
from .config import AvoidCharacterGroup as ConfigAvoidCharacterGroup
from .config import (
    CollectionConfig,
    Config,
    DevelLikeBranch,
    PackageType,
    Sessions,
)
from .paths.match import FileCollector
from .sessions.ansible_lint import add_ansible_lint
from .sessions.ansible_test import (
    AnsibleTestIntegrationSessionTemplate,
    AnsibleTestIntegrationSessionTemplateGroup,
    add_all_ansible_test_sanity_test_sessions,
    add_all_ansible_test_unit_test_sessions,
    add_ansible_test_integration_sessions,
    add_ansible_test_integration_sessions_default_container,
)
from .sessions.build_import_check import add_build_import_check
from .sessions.docs_check import add_docs_check
from .sessions.ee_check import (
    ExecutionEnvironmentData,
    add_execution_environment_sessions,
)
from .sessions.extra_checks import (
    ActionGroup,
    AvoidCharacterGroup,
    add_extra_checks,
)
from .sessions.license_check import add_license_check
from .sessions.lint import add_lint_sessions
from .sessions.matrix_generator import add_matrix_generator
from .sessions.utils.packages import PackageType as RuntimePackageType
from .utils import Version

_T = t.TypeVar("_T", bound=object)


def _interpret_config(config: Config) -> None:
    if config.collection_sources:
        setup_collection_sources(
            {
                name: CollectionSource(name=name, source=source.source)
                for name, source in config.collection_sources.items()
            }
        )
    if config.collection_sources_per_ansible:
        for (
            ansible_core_version,
            collection_sources,
        ) in config.collection_sources_per_ansible.items():
            setup_collection_sources(
                {
                    name: CollectionSource(name=name, source=source.source)
                    for name, source in collection_sources.items()
                },
                ansible_core_version=ansible_core_version,
            )


def _convert_action_groups(
    action_groups: list[ConfigActionGroup] | None,
) -> list[ActionGroup] | None:
    if action_groups is None:
        return None
    return [
        ActionGroup(
            name=action_group.name,
            pattern=action_group.pattern,
            doc_fragment=action_group.doc_fragment,
            exclusions=action_group.exclusions,
        )
        for action_group in action_groups
    ]


def _convert_avoid_character_groups(
    avoid_character_groups: list[ConfigAvoidCharacterGroup] | None,
) -> list[AvoidCharacterGroup] | None:
    if avoid_character_groups is None:
        return None
    return [
        AvoidCharacterGroup(
            name=avoid_character_group.name,
            regex=avoid_character_group.regex,
            match_extensions=avoid_character_group.match_extensions,
            match_paths=avoid_character_group.match_paths,
            match_directories=avoid_character_group.match_directories,
            skip_extensions=avoid_character_group.skip_extensions,
            skip_paths=avoid_character_group.skip_paths,
            skip_directories=avoid_character_group.skip_directories,
        )
        for avoid_character_group in avoid_character_groups
    ]


def _convert_devel_like_branch(
    devel_like_branch: DevelLikeBranch | None,
) -> tuple[str | None, str] | None:
    if devel_like_branch is None:
        return None
    return devel_like_branch.repository, devel_like_branch.branch


def _convert_devel_like_branches(
    devel_like_branches: list[DevelLikeBranch] | None,
) -> list[tuple[str | None, str]] | None:
    if devel_like_branches is None:
        return None
    return [(branch.repository, branch.branch) for branch in devel_like_branches]


def _convert_except_versions(
    except_versions: list[AnsibleCoreVersion] | None,
) -> list[AnsibleCoreVersion | str] | None:
    return t.cast(t.Optional[list[t.Union[AnsibleCoreVersion, str]]], except_versions)


def _convert_core_python_versions(
    core_python_versions: dict[AnsibleCoreVersion | str, list[Version]] | None,
) -> dict[str | AnsibleCoreVersion, list[str | Version]] | None:
    return t.cast(
        t.Optional[dict[t.Union[str, AnsibleCoreVersion], list[t.Union[str, Version]]]],
        core_python_versions,
    )


def _ensure_list(value: _T | list[_T]) -> list[_T]:
    if isinstance(value, list):
        return value
    return [value]


def _ensure_list_w_none(value: _T | list[_T] | None) -> list[_T | None]:
    if isinstance(value, list):
        return value  # type: ignore
    return [value]


@t.overload
def _convert_package_name(
    package_name: PackageType | list[PackageType],
) -> list[RuntimePackageType]: ...


@t.overload
def _convert_package_name(
    package_name: list[PackageType | str],
) -> list[RuntimePackageType | str]: ...


@t.overload
def _convert_package_name(
    package_name: PackageType | list[PackageType] | None,
) -> list[RuntimePackageType] | None: ...


def _convert_package_name(
    package_name: (
        PackageType | str | list[PackageType] | list[PackageType | str] | None
    ),
) -> list[RuntimePackageType | str] | None:
    if package_name is None:
        return None
    package_name_list: list[PackageType] | list[PackageType | str] = _ensure_list(
        package_name
    )  # type: ignore
    return [
        pkg if isinstance(pkg, str) else pkg.to_utils_instance()
        for pkg in package_name_list
    ]


def _coalesce(*values: _T) -> _T | None:
    for value in values:
        if value is not None:
            return value
    return None


def _add_ansible_test_sessions(sessions: Sessions, cconfig: CollectionConfig) -> None:
    if sessions.ansible_test_sanity:
        add_all_ansible_test_sanity_test_sessions(
            default=sessions.ansible_test_sanity.default,
            include_devel=sessions.ansible_test_sanity.include_devel,
            include_milestone=sessions.ansible_test_sanity.include_milestone,
            add_devel_like_branches=_convert_devel_like_branches(
                sessions.ansible_test_sanity.add_devel_like_branches
            ),
            min_version=sessions.ansible_test_sanity.min_version,
            max_version=sessions.ansible_test_sanity.max_version,
            except_versions=_convert_except_versions(
                sessions.ansible_test_sanity.except_versions
            ),
            skip_tests=sessions.ansible_test_sanity.skip_tests,
            allow_disabled=sessions.ansible_test_sanity.allow_disabled,
            enable_optional_errors=sessions.ansible_test_sanity.enable_optional_errors,
        )
    if sessions.ansible_test_units:
        add_all_ansible_test_unit_test_sessions(
            default=sessions.ansible_test_units.default,
            include_devel=sessions.ansible_test_units.include_devel,
            include_milestone=sessions.ansible_test_units.include_milestone,
            add_devel_like_branches=_convert_devel_like_branches(
                sessions.ansible_test_units.add_devel_like_branches
            ),
            min_version=sessions.ansible_test_units.min_version,
            max_version=sessions.ansible_test_units.max_version,
            except_versions=_convert_except_versions(
                sessions.ansible_test_units.except_versions
            ),
        )

    # Handle integration tests
    integration_defaults: list[bool] = []
    if sessions.ansible_test_integration_w_default_container:
        integration_defaults.append(
            sessions.ansible_test_integration_w_default_container.default
        )
    if sessions.ansible_test_integration:
        integration_defaults.append(sessions.ansible_test_integration.default)

    default_per_int_session = len(set(integration_defaults)) == 2

    integration_sessions: list[str] = []
    if sessions.ansible_test_integration_w_default_container:
        cfg = sessions.ansible_test_integration_w_default_container
        integration_sessions.extend(
            add_ansible_test_integration_sessions_default_container(
                default=default_per_int_session and cfg.default,
                include_devel=cfg.include_devel,
                include_milestone=(cfg.include_milestone),
                add_devel_like_branches=_convert_devel_like_branches(
                    cfg.add_devel_like_branches
                ),
                min_version=cfg.min_version,
                max_version=cfg.max_version,
                except_versions=_convert_except_versions(cfg.except_versions),
                core_python_versions=_convert_core_python_versions(
                    cfg.core_python_versions
                ),
                controller_python_versions_only=cfg.controller_python_versions_only,
                min_python_version=cconfig.min_python_version,
                ansible_vars_from_env_vars=cfg.ansible_vars_from_env_vars,
                ansible_vars={
                    k: v.to_utils_instance() for k, v in cfg.ansible_vars.items()
                },
            )
        )
    if sessions.ansible_test_integration:
        integration_sessions.extend(
            add_ansible_test_integration_sessions(
                default=default_per_int_session
                and sessions.ansible_test_integration.default,
                ansible_vars={
                    k: v.to_utils_instance()
                    for k, v in sessions.ansible_test_integration.ansible_vars.items()
                },
                global_tags=sessions.ansible_test_integration.tags,
                session_templates=[
                    AnsibleTestIntegrationSessionTemplate(
                        ansible_core=_ensure_list(session.ansible_core),
                        docker=_ensure_list_w_none(session.docker),
                        remote=_ensure_list_w_none(session.remote),
                        python_version=_ensure_list_w_none(session.python_version),
                        target=_ensure_list_w_none(session.target),
                        gha_container=_ensure_list_w_none(session.gha_container),
                        devel_like_branch=_convert_devel_like_branch(
                            session.devel_like_branch
                        ),
                        ansible_vars={
                            k: v.to_utils_instance()
                            for k, v in session.ansible_vars.items()
                        },
                        session_name_template=session.session_name_template
                        or sessions.ansible_test_integration.session_name_template,
                        display_name_template=session.display_name_template
                        or sessions.ansible_test_integration.display_name_template,
                        description_template=session.description_template
                        or sessions.ansible_test_integration.description_template,
                        tags=session.tags,
                    )
                    for session in sessions.ansible_test_integration.sessions
                ],
                session_template_groups=[
                    AnsibleTestIntegrationSessionTemplateGroup(
                        session_name=group.session_name,
                        description=group.description,
                        ansible_vars={
                            k: v.to_utils_instance()
                            for k, v in group.ansible_vars.items()
                        },
                        tags=group.tags,
                        session_templates=[
                            AnsibleTestIntegrationSessionTemplate(
                                ansible_core=_ensure_list(session.ansible_core),
                                docker=_ensure_list_w_none(
                                    _coalesce(session.docker, group.docker)
                                ),
                                remote=_ensure_list_w_none(
                                    _coalesce(session.remote, group.remote)
                                ),
                                python_version=_ensure_list_w_none(
                                    _coalesce(
                                        session.python_version, group.python_version
                                    )
                                ),
                                target=_ensure_list_w_none(
                                    _coalesce(session.target, group.target)
                                ),
                                gha_container=_ensure_list_w_none(
                                    _coalesce(
                                        session.gha_container, group.gha_container
                                    )
                                ),
                                devel_like_branch=_convert_devel_like_branch(
                                    session.devel_like_branch
                                ),
                                ansible_vars={
                                    k: v.to_utils_instance()
                                    for k, v in session.ansible_vars.items()
                                },
                                session_name_template=session.session_name_template
                                or group.session_name_template
                                or sessions.ansible_test_integration.session_name_template,
                                display_name_template=session.display_name_template
                                or group.display_name_template
                                or sessions.ansible_test_integration.display_name_template,
                                description_template=session.description_template
                                or group.description_template
                                or sessions.ansible_test_integration.description_template,
                                tags=session.tags,
                            )
                            for session in group.sessions
                        ],
                    )
                    for group in sessions.ansible_test_integration.groups
                ],
            )
        )

    if integration_defaults:
        duplicates_check: set[str] = set()
        for name in integration_sessions:
            # nox does not warn/error on duplicate session names
            # (https://github.com/wntrblm/nox/issues/998)
            if name in duplicates_check:
                raise ValueError(
                    f"Two integration test sessions with name {name} were added!"
                )
            duplicates_check.add(name)

        def ansible_test_integration(
            session: nox.Session,  # pylint: disable=unused-argument
        ) -> None:
            pass

        ansible_test_integration.__doc__ = (
            "Meta session for running all ansible-test-integration-* sessions."
        )
        nox.session(
            name="ansible-test-integration",
            requires=integration_sessions,
            default=False if default_per_int_session else integration_defaults[0],
        )(ansible_test_integration)


def _add_sessions(sessions: Sessions, cconfig: CollectionConfig) -> None:
    if sessions.lint:
        code_files: FileCollector | None = None
        if sessions.lint.code_files != "default":
            code_files = FileCollector.create(sessions.lint.code_files, glob=True)
        module_files: FileCollector | None = None
        if sessions.lint.module_files != "default":
            module_files = FileCollector.create(sessions.lint.module_files, glob=True)
        add_lint_sessions(
            make_lint_default=sessions.lint.default,
            code_files=code_files,
            module_files=module_files,
            extra_code_files=sessions.lint.extra_code_files,
            run_isort=sessions.lint.run_isort,
            isort_config=sessions.lint.isort_config,
            isort_modules_config=sessions.lint.isort_modules_config,
            isort_package=_convert_package_name(sessions.lint.isort_package),
            run_black=sessions.lint.run_black,
            run_black_modules=sessions.lint.run_black_modules,
            black_config=sessions.lint.black_config,
            black_modules_config=sessions.lint.black_modules_config,
            black_package=_convert_package_name(sessions.lint.black_package),
            run_ruff_format=sessions.lint.run_ruff_format,
            ruff_format_config=sessions.lint.ruff_format_config
            or sessions.lint.ruff_config,
            ruff_format_modules_config=sessions.lint.ruff_format_modules_config
            or sessions.lint.ruff_modules_config,
            ruff_format_package=_convert_package_name(
                sessions.lint.ruff_format_package or sessions.lint.ruff_package
            ),
            run_ruff_autofix=sessions.lint.run_ruff_autofix,
            ruff_autofix_config=sessions.lint.ruff_autofix_config
            or sessions.lint.ruff_config,
            ruff_autofix_modules_config=sessions.lint.ruff_autofix_modules_config
            or sessions.lint.ruff_modules_config,
            ruff_autofix_package=_convert_package_name(
                sessions.lint.ruff_autofix_package or sessions.lint.ruff_package
            ),
            ruff_autofix_select=sessions.lint.ruff_autofix_select,
            run_ruff_check=sessions.lint.run_ruff_check,
            ruff_check_config=sessions.lint.ruff_check_config
            or sessions.lint.ruff_config,
            ruff_check_modules_config=sessions.lint.ruff_check_modules_config
            or sessions.lint.ruff_modules_config,
            ruff_check_package=_convert_package_name(
                sessions.lint.ruff_check_package or sessions.lint.ruff_package
            ),
            run_flake8=sessions.lint.run_flake8,
            flake8_config=sessions.lint.flake8_config,
            flake8_modules_config=sessions.lint.flake8_modules_config,
            flake8_package=_convert_package_name(sessions.lint.flake8_package),
            run_pylint=sessions.lint.run_pylint,
            pylint_rcfile=sessions.lint.pylint_rcfile,
            pylint_modules_rcfile=sessions.lint.pylint_modules_rcfile,
            pylint_package=_convert_package_name(sessions.lint.pylint_package),
            pylint_ansible_core_package=_convert_package_name(
                sessions.lint.pylint_ansible_core_package
            ),
            pylint_extra_deps=_convert_package_name(sessions.lint.pylint_extra_deps),
            run_yamllint=sessions.lint.run_yamllint,
            yamllint_config=sessions.lint.yamllint_config,
            yamllint_config_plugins=sessions.lint.yamllint_config_plugins,
            yamllint_config_plugins_examples=sessions.lint.yamllint_config_plugins_examples,
            yamllint_config_extra_docs=sessions.lint.yamllint_config_extra_docs,
            yamllint_package=_convert_package_name(sessions.lint.yamllint_package),
            yamllint_antsibull_docutils_package=_convert_package_name(
                sessions.lint.yamllint_antsibull_docutils_package
            ),
            run_mypy=sessions.lint.run_mypy,
            mypy_config=sessions.lint.mypy_config,
            mypy_modules_config=sessions.lint.mypy_modules_config,
            mypy_package=_convert_package_name(sessions.lint.mypy_package),
            mypy_ansible_core_package=_convert_package_name(
                sessions.lint.mypy_ansible_core_package
            ),
            mypy_extra_deps=_convert_package_name(sessions.lint.mypy_extra_deps),
            run_antsibullnox_config_lint=sessions.lint.run_antsibullnox_config_lint,
        )
    if sessions.docs_check:
        add_docs_check(
            make_docs_check_default=sessions.docs_check.default,
            antsibull_docs_package=_convert_package_name(
                sessions.docs_check.antsibull_docs_package
            ),
            ansible_core_package=_convert_package_name(
                sessions.docs_check.ansible_core_package
            ),
            validate_collection_refs=sessions.docs_check.validate_collection_refs,
            extra_collections=sessions.docs_check.extra_collections,
            codeblocks_restrict_types=sessions.docs_check.codeblocks_restrict_types,
            codeblocks_restrict_type_exact_case=(
                sessions.docs_check.codeblocks_restrict_type_exact_case
            ),
            codeblocks_allow_without_type=sessions.docs_check.codeblocks_allow_without_type,
            codeblocks_allow_literal_blocks=sessions.docs_check.codeblocks_allow_literal_blocks,
            antsibull_docutils_package=_convert_package_name(
                sessions.docs_check.antsibull_docutils_package
            ),
        )
    if sessions.license_check:
        add_license_check(
            make_license_check_default=sessions.license_check.default,
            run_reuse=sessions.license_check.run_reuse,
            reuse_package=_convert_package_name(sessions.license_check.reuse_package),
            run_license_check=sessions.license_check.run_license_check,
            license_check_extra_ignore_paths=(
                sessions.license_check.license_check_extra_ignore_paths
            ),
        )
    if sessions.extra_checks:
        add_extra_checks(
            make_extra_checks_default=sessions.extra_checks.default,
            run_no_unwanted_files=sessions.extra_checks.run_no_unwanted_files,
            no_unwanted_files_module_extensions=(
                sessions.extra_checks.no_unwanted_files_module_extensions
            ),
            no_unwanted_files_other_extensions=(
                sessions.extra_checks.no_unwanted_files_other_extensions
            ),
            no_unwanted_files_yaml_extensions=(
                sessions.extra_checks.no_unwanted_files_yaml_extensions
            ),
            no_unwanted_files_skip_paths=(
                sessions.extra_checks.no_unwanted_files_skip_paths
            ),
            no_unwanted_files_skip_directories=(
                sessions.extra_checks.no_unwanted_files_skip_directories
            ),
            no_unwanted_files_yaml_directories=(
                sessions.extra_checks.no_unwanted_files_yaml_directories
            ),
            no_unwanted_files_allow_symlinks=(
                sessions.extra_checks.no_unwanted_files_allow_symlinks
            ),
            run_action_groups=sessions.extra_checks.run_action_groups,
            action_groups_config=_convert_action_groups(
                sessions.extra_checks.action_groups_config
            ),
            run_no_trailing_whitespace=sessions.extra_checks.run_no_trailing_whitespace,
            no_trailing_whitespace_skip_paths=(
                sessions.extra_checks.no_trailing_whitespace_skip_paths
            ),
            no_trailing_whitespace_skip_directories=(
                sessions.extra_checks.no_trailing_whitespace_skip_directories
            ),
            run_avoid_characters=sessions.extra_checks.run_avoid_characters,
            avoid_character_group=_convert_avoid_character_groups(
                sessions.extra_checks.avoid_character_group
            ),
        )
    if sessions.build_import_check:
        add_build_import_check(
            make_build_import_check_default=sessions.build_import_check.default,
            ansible_core_package=_convert_package_name(
                sessions.build_import_check.ansible_core_package
            ),
            run_galaxy_importer=sessions.build_import_check.run_galaxy_importer,
            galaxy_importer_package=_convert_package_name(
                sessions.build_import_check.galaxy_importer_package
            ),
            galaxy_importer_config_path=sessions.build_import_check.galaxy_importer_config_path,
            galaxy_importer_always_show_logs=(
                sessions.build_import_check.galaxy_importer_always_show_logs
            ),
        )
    _add_ansible_test_sessions(sessions, cconfig)
    if sessions.ansible_lint:
        add_ansible_lint(
            make_ansible_lint_default=sessions.ansible_lint.default,
            ansible_lint_package=_convert_package_name(
                sessions.ansible_lint.ansible_lint_package
            ),
            additional_requirements_files=sessions.ansible_lint.additional_requirements_files,
            strict=sessions.ansible_lint.strict,
        )
    if sessions.ee_check:
        execution_environments = []

        for ee_config in sessions.ee_check.execution_environments:
            execution_environments.append(
                ExecutionEnvironmentData(
                    name=ee_config.name,
                    description=ee_config.description or ee_config.name,
                    config=ee_config.to_execution_environment_config(),
                    test_playbooks=ee_config.test_playbooks,
                    runtime_environment=ee_config.runtime_environment,
                    runtime_container_options=ee_config.runtime_container_options,
                    runtime_extra_vars=ee_config.runtime_extra_vars,
                )
            )

        if execution_environments:
            add_execution_environment_sessions(
                execution_environments=execution_environments,
                default=sessions.ee_check.default,
                ansible_builder_package=_convert_package_name(
                    sessions.ee_check.ansible_builder_package
                ),
                ansible_core_package=_convert_package_name(
                    sessions.ee_check.ansible_core_package
                ),
                ansible_navigator_package=_convert_package_name(
                    sessions.ee_check.ansible_navigator_package
                ),
            )
    add_matrix_generator()


def interpret_config(config: Config) -> None:
    """
    Interpret the config file's contents.
    """
    _interpret_config(config)
    _add_sessions(config.sessions, config.collection)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2026, Ansible Project

# pylint: disable=missing-function-docstring

"""
Test path matchers.
"""

from __future__ import annotations

from pathlib import Path

import pytest

from antsibull_nox.paths.match import FileCollector, _FileSet, _FileTree

from ..utils import chdir


def test_FileSet() -> None:
    fs1 = _FileSet.create([Path("foo")])
    fs2 = _FileSet.create([Path("bar")])
    fs3 = _FileSet.create([Path("foo"), Path("bar")])

    fs = _FileSet.create_from_split(split_files=[("foo",), ("foo",)])
    assert fs == fs1

    assert fs3.subset({("foo",)}) == fs1
    assert fs3.subset({("bar",)}) == fs2

    fs = fs1.clone()
    fs.merge_set(fs1)
    assert fs == fs1

    fs = fs1.clone()
    fs.merge_set(fs2)
    assert fs == fs3

    fs = fs1.clone()
    fs.merge_paths(paths=[Path("foo")])
    assert fs == fs1

    fs = fs1.clone()
    fs.merge_paths(paths=[Path("bar")])
    assert fs == fs3


def test_FileTree() -> None:
    ft = _FileTree.create_from_paths(
        [
            Path("foo/bar"),
            Path("foo"),
            Path("foo/bam"),
            Path("bam/baz/bar"),
        ],
        keep_pruned=True,
    )
    assert sorted(ft.iterate()) == [("bam", "baz", "bar"), ("foo",)]

    assert ft.has_or_has_children(())
    assert not ft.has_or_has_children(("bar",))
    assert ft.has_or_has_children(("foo",))
    assert not ft.has_or_has_children(("foo", "bar"))
    assert not ft.has_or_has_children(("foo", "baz"))
    assert ft.has_or_has_children(("bam",))
    assert not ft.has_or_has_children(("bam", "bar"))
    assert ft.has_or_has_children(("bam", "baz"))
    assert ft.has_or_has_children(("bam", "baz", "bar"))
    assert not ft.has_or_has_children(("bam", "baz", "bar", "boom"))

    assert not ft.has_or_is_child(())
    assert not ft.has_or_is_child(("bar",))
    assert ft.has_or_is_child(("foo",))
    assert ft.has_or_is_child(("foo", "bar"))
    assert ft.has_or_is_child(("foo", "baz"))
    assert not ft.has_or_is_child(("bam",))
    assert not ft.has_or_is_child(("bam", "bar"))
    assert not ft.has_or_is_child(("bam", "baz"))
    assert ft.has_or_is_child(("bam", "baz", "bar"))
    assert ft.has_or_is_child(("bam", "baz", "bar", "boom"))

    assert sorted(ft.iterate(prefix=("foo", "foo"))) == []

    ft = _FileTree.create_from_paths(
        [
            Path("foo/bar"),
            Path("foo"),
            Path("foo/bam"),
            Path("."),
        ],
        keep_pruned=False,
    )
    assert sorted(ft.iterate()) == [(), ("foo",), ("foo", "bam"), ("foo", "bar")]


@pytest.fixture(name="path_structure", scope="module")
def create_path_structure(tmp_path_factory: pytest.TempPathFactory) -> Path:
    root = tmp_path_factory.mktemp("path-structure")
    (root / "a").write_text("")
    (root / "b").write_text("")

    c = root / "c"
    c.mkdir()
    (c / "ca").write_text("")
    (c / "cb").write_text("")
    cpycache = c / "__pycache__"
    cpycache.mkdir()
    (cpycache / "x").write_text("")
    (c / "ca").write_text("")
    (c / "cb").write_text("")
    cc = c / "cc"
    cc.mkdir()
    (cc / "cca").write_text("")
    (cc / "ccb").write_text("")
    cd = c / "cd"
    cd.mkdir()
    (cd / "cda").write_text("")
    (cd / "cdb").write_text("")

    d = root / "d"
    d.mkdir()
    (d / "da").write_text("")
    (d / "db").write_text("")
    dc = d / "dc"
    dc.mkdir()
    (dc / "dca").write_text("")
    (dc / "dcb").write_text("")
    dd = d / "dd"
    dd.mkdir()
    (dd / "dda").write_text("")
    (dd / "ddb").write_text("")

    e = root / "e"
    e.mkdir()
    (e / "ea.py").write_text("")
    (e / "eb.txt").write_text("")
    ec = e / "ec"
    ec.mkdir()
    (ec / "eca.py").write_text("")
    (ec / "ecb.txt").write_text("")
    ed = e / "ed.py"
    ed.mkdir()
    (ed / "eda.py").write_text("")
    (ed / "edb.txt").write_text("")
    ee = e / "ee.txt"
    ee.mkdir()
    (ee / "eea.py").write_text("")
    (ee / "eeb.txt").write_text("")

    return root


TEST_GET_PATHS: list[tuple[list[str], list[str]]] = [
    (
        ["."],
        ["."],
    ),
    (
        [".", "."],
        ["."],
    ),
    (
        ["foo", "bar", "baz", "bar/"],
        ["foo", "bar", "baz"],
    ),
]


@pytest.mark.parametrize(
    "start_paths, expected_paths",
    TEST_GET_PATHS,
)
def test_get_paths(
    start_paths: list[str],
    expected_paths: list[str],
) -> None:
    fc = FileCollector(paths=[Path(path) for path in start_paths])
    result = [str(path) for path in fc.get_paths()]
    assert result == expected_paths

    assert fc.clone().get_paths() == fc.get_paths()


TEST_GET_EXISTING: list[tuple[list[str], list[str]]] = [
    (
        ["."],
        ["."],
    ),
    (
        ["c/cd/cda"],
        ["c/cd/cda"],
    ),
    (
        ["foo", "bar", "baz", "a", "c", "c/a", "c/ca"],
        ["a", "c", "c/ca"],
    ),
]


@pytest.mark.parametrize(
    "start_paths, expected_paths",
    TEST_GET_EXISTING,
)
def test_get_existing(
    start_paths: list[str],
    expected_paths: list[str],
    path_structure: Path,
) -> None:
    with chdir(path_structure):
        fc = FileCollector(paths=[Path(path) for path in start_paths])
        result = [str(path) for path in fc.get_existing()]
        assert result == expected_paths


TEST_RESTRICT: list[tuple[list[str], list[str], list[str]]] = [
    (
        ["."],
        ["."],
        ["."],
    ),
    (
        ["."],
        ["foo"],
        ["foo"],
    ),
    (
        ["foo/bar"],
        ["foo"],
        ["foo/bar"],
    ),
    (
        ["bam", "bam/foo/bar", "bar"],
        ["bam/foo"],
        ["bam/foo", "bam/foo/bar"],
    ),
    (
        ["."],
        ["foo/bar", "foo", "foo/bam"],
        ["foo"],
    ),
    (
        ["foo", "bar"],
        ["foo", "bar/baz", "bar/bam", "bam"],
        ["bar/bam", "bar/baz", "foo"],
    ),
    (
        ["foo/bam", "foo/bar"],
        ["foo"],
        ["foo/bam", "foo/bar"],
    ),
    (
        ["foo"],
        ["foo/bam", "foo/bar"],
        ["foo/bam", "foo/bar"],
    ),
    (
        ["foo"],
        ["foo"],
        ["foo"],
    ),
]


@pytest.mark.parametrize(
    "start_paths, restrict_paths, expected_paths",
    TEST_RESTRICT,
)
def test_restrict(
    start_paths: list[str],
    restrict_paths: list[str],
    expected_paths: list[str],
) -> None:
    fc = FileCollector(paths=[Path(path) for path in start_paths])
    fc.restrict(paths=[Path(path) for path in restrict_paths])
    result = sorted(str(path) for path in fc.get_paths())
    assert result == expected_paths

    # Alternative way
    fc1 = FileCollector(paths=[Path(path) for path in start_paths])
    fc2 = FileCollector(paths=[Path(path) for path in restrict_paths])
    fc1.restrict(paths=fc2)
    result = sorted(str(path) for path in fc1.get_paths())
    assert result == expected_paths


TEST_REMOVE: list[tuple[list[str], list[str], list[str] | None, list[str]]] = [
    (
        ["c"],
        ["."],
        None,
        [],
    ),
    (
        ["a", "b", "c"],
        ["d"],
        None,
        ["a", "b", "c"],
    ),
    (
        ["a", "b", "c", "d", "d/da", "d/db", "d/dc/dca"],
        ["c/a", "d"],
        None,
        ["a", "b", "c/ca", "c/cb", "c/cc", "c/cd"],
    ),
    (
        ["a", "b", "c", "d", "d/da", "d/db", "d/dc/dca"],
        ["c/ca", "c/cc", "c/cd/cda", "d/dc"],
        None,
        ["a", "b", "c/cb", "c/cd/cdb", "d/da", "d/db", "d/dd"],
    ),
    (
        ["d"],
        ["d/da", "d/dc"],
        None,
        ["d/db", "d/dd"],
    ),
    (
        ["e"],
        ["e/x"],
        ["py", ".txt"],
        ["e/ea.py", "e/ec", "e/ed.py", "e/ee.txt"],
    ),
]


@pytest.mark.parametrize(
    "start_paths, remove_paths, extensions, expected_paths",
    TEST_REMOVE,
)
def test_remove(
    start_paths: list[str],
    remove_paths: list[str],
    extensions: list[str] | None,
    expected_paths: list[str],
    path_structure: Path,
) -> None:
    with chdir(path_structure):
        fc = FileCollector(paths=[Path(path) for path in start_paths])
        fc.remove(paths=[Path(path) for path in remove_paths], extensions=extensions)
        result = sorted(str(path) for path in fc.get_paths())
        assert result == expected_paths


TEST_CREATE: list[tuple[list[str], bool, list[str]]] = [
    (
        ["c", "moo"],
        False,
        ["c", "moo"],
    ),
    (
        ["*/?c", "*/*fff", "e", "moo"],
        True,
        ["c/cc", "d/dc", "e", "moo"],
    ),
    (
        ["."],
        True,
        ["."],
    ),
]


@pytest.mark.parametrize(
    "sources, glob, expected_paths",
    TEST_CREATE,
)
def test_create(
    sources: list[str],
    glob: bool,
    expected_paths: list[str],
    path_structure: Path,
) -> None:
    with chdir(path_structure):
        fc = FileCollector.create(sources, glob=glob)
        result = sorted(str(path) for path in fc.get_paths())
        assert result == expected_paths



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox lint sessions.
"""

from __future__ import annotations

import os
import shlex
from pathlib import Path

import nox

from ..collection import (
    load_collection_data_from_disk,
)
from ..messages import Message
from ..messages.parse import (
    parse_mypy_errors,
    parse_pylint_json2_errors,
    parse_ruff_check_errors,
)
from ..paths.match import (
    FileCollector,
)
from ..paths.utils import (
    list_all_files,
    relative_to_walk_up,
)
from .collections import (
    CollectionSetup,
    prepare_collections,
)
from .docs_check import find_extra_docs_rst_files
from .utils import (
    IN_CI,
    compose_description,
    silence_run_verbosity,
)
from .utils.output import (
    print_messages,
)
from .utils.packages import (
    PackageType,
    PackageTypeOrList,
    check_package_types,
    install,
    normalize_package_type,
)
from .utils.paths import (
    PythonDependencies,
    filter_files_cd,
    filter_paths,
)
from .utils.scripts import (
    run_bare_script,
)

CODE_FILES = [
    Path("plugins"),
    Path("tests/unit"),
]

CODE_FILES_W_NOXFILE = [
    Path("plugins"),
    Path("tests/unit"),
    Path("noxfile.py"),
]

MODULE_PATHS = FileCollector.create(
    [
        "plugins/modules/",
        "plugins/module_utils/",
        "tests/unit/plugins/modules/",
        "tests/unit/plugins/module_utils/",
    ],
    glob=False,
)


def _split_arg(
    session: nox.Session, arg: str | PackageType, arg_name: str, index: int
) -> list[str | PackageType]:
    if not isinstance(arg, str):
        return check_package_types(session, arg_name, [arg])
    args = shlex.split(arg)
    # How to resolve the deprecations:
    #   1. Eventually, make the following warnings errors (in 2.0.0 likely);
    #   2. Eventually, parse str as PackageName as for other
    #      package list arguments (do this in config.py) (in 2.x.0 likely).
    if args != [arg]:
        session.warn(
            f"DEPRECATION WARNING: {arg_name}[{index + 1}] is currently shell-split."
            " This behavior is deprecated and will change in a future release."
            " Specify the dependency as a dictionary with 'type' to avoid ambiguity;"
            " see PackageType in the config file documentation for details."
        )
    for part in args:
        if part.startswith("-"):
            session.warn(
                f"DEPRECATION WARNING: {arg_name}[{index + 1}] contains an argument"
                f" {part!r} starting with a dash."
                " This behavior is deprecated and will stop working in a future release."
                " Specify this dependency as a package type dictionary;"
                " see PackageType in the config file documentation for details."
            )
    return args  # type: ignore


def add_lint(
    *,
    make_lint_default: bool,
    has_formatters: bool,
    has_codeqa: bool,
    has_yamllint: bool,
    has_typing: bool,
    has_config_lint: bool,
) -> None:
    """
    Add nox meta session for linting.
    """

    def lint(session: nox.Session) -> None:  # pylint: disable=unused-argument
        pass  # this session is deliberately empty

    dependent_sessions = []
    if has_formatters:
        dependent_sessions.append("formatters")
    if has_codeqa:
        dependent_sessions.append("codeqa")
    if has_yamllint:
        dependent_sessions.append("yamllint")
    if has_typing:
        dependent_sessions.append("typing")
    if has_config_lint:
        dependent_sessions.append("antsibull-nox-config")

    lint.__doc__ = compose_description(
        prefix={
            "one": "Meta session for triggering the following session:",
            "other": "Meta session for triggering the following sessions:",
        },
        programs={
            "formatters": has_formatters,
            "codeqa": has_codeqa,
            "yamllint": has_yamllint,
            "typing": has_typing,
            "antsibull-nox-config": has_config_lint,
        },
    )
    nox.session(
        name="lint",
        default=make_lint_default,
        requires=dependent_sessions,
    )(lint)


def _get_files(
    *,
    code_files: list[Path] | FileCollector,
    module_files: list[Path] | FileCollector,
    split_modules: bool,
    cd_add_python_deps: PythonDependencies = "none",
) -> tuple[list[Path] | None, list[Path] | None, list[Path] | None]:
    files: list[Path] | None = None
    files_modules: list[Path] | None = None
    files_other: list[Path] | None = None
    if split_modules:
        files_modules = filter_paths(
            code_files,
            restrict=module_files,
            extensions=[".py"],
            with_cd=True,
            cd_add_python_deps=cd_add_python_deps,
        )
        files_other = filter_paths(
            code_files,
            remove=module_files,
            extensions=[".py"],
            with_cd=True,
            cd_add_python_deps=cd_add_python_deps,
        )
    else:
        files = filter_paths(
            code_files,
            extensions=[".py"],
            with_cd=True,
            cd_add_python_deps=cd_add_python_deps,
        )
    return files, files_modules, files_other


def _execute_isort_for(
    session: nox.Session,
    *,
    old_cwd: Path,
    root_dir: Path,
    collection_dir: Path,
    run_check: bool,
    paths: list[Path],
    isort_config: str | os.PathLike | None,
    what_for: str = "",
) -> None:
    if not paths:
        session.warn(f"Skipping isort{what_for} (no files to process)")
        return
    command: list[str] = [
        "isort",
        "--src",
        ".",
    ]
    if run_check:
        command.append("--check")
    if isort_config is not None:
        command.extend(
            [
                "--settings-file",
                str(relative_to_walk_up((old_cwd / isort_config).resolve(), root_dir)),
            ]
        )
    command.extend(session.posargs)
    relative_dir = collection_dir.relative_to(root_dir)
    for file in paths:
        command.append(str(relative_dir / file))
    session.run(*command)


def _execute_isort(
    session: nox.Session,
    *,
    root_dir: Path,
    collection_dir: Path,
    run_check: bool,
    code_files: list[Path] | FileCollector,
    module_files: list[Path] | FileCollector,
    isort_config: str | os.PathLike | None,
    isort_modules_config: str | os.PathLike | None,
) -> None:
    files, files_modules, files_other = _get_files(
        code_files=code_files,
        module_files=module_files,
        split_modules=isort_modules_config is not None
        and isort_modules_config != isort_config,
    )
    old_cwd = Path.cwd()
    with session.chdir(root_dir):
        if files is not None:
            _execute_isort_for(
                session,
                old_cwd=old_cwd,
                root_dir=root_dir,
                collection_dir=collection_dir,
                run_check=run_check,
                paths=files,
                isort_config=isort_config,
            )
        if files_modules is not None:
            _execute_isort_for(
                session,
                old_cwd=old_cwd,
                root_dir=root_dir,
                collection_dir=collection_dir,
                run_check=run_check,
                paths=files_modules,
                isort_config=isort_modules_config or isort_config,
                what_for=" for modules and module utils",
            )
        if files_other is not None:
            _execute_isort_for(
                session,
                old_cwd=old_cwd,
                root_dir=root_dir,
                collection_dir=collection_dir,
                run_check=run_check,
                paths=files_other,
                isort_config=isort_config,
                what_for=" for other files",
            )


def _execute_black_for(
    session: nox.Session,
    *,
    paths: list[Path],
    run_check: bool,
    black_config: str | os.PathLike | None,
    what_for: str = "",
) -> None:
    if not paths:
        session.warn(f"Skipping black{what_for} (no files to process)")
        return
    command = ["black"]
    if run_check:
        command.append("--check")
    if black_config is not None:
        command.extend(["--config", str(black_config)])
    command.extend(session.posargs)
    command.extend(str(path) for path in paths)
    session.run(*command)


def _execute_black(
    session: nox.Session,
    *,
    run_check: bool,
    code_files: list[Path] | FileCollector,
    module_files: list[Path] | FileCollector,
    run_black: bool,
    run_black_modules: bool | None,
    black_config: str | os.PathLike | None,
    black_modules_config: str | os.PathLike | None,
) -> None:
    if (
        run_black
        and run_black_modules
        and (black_modules_config is None or black_modules_config == black_config)
    ):
        _execute_black_for(
            session,
            paths=filter_paths(
                code_files,
                extensions=[".py"],
                with_cd=True,
            ),
            run_check=run_check,
            black_config=black_config,
        )
        return
    if run_black:
        paths = filter_paths(
            code_files,
            remove=module_files,
            extensions=[".py"],
            with_cd=True,
        )
        _execute_black_for(
            session,
            paths=paths,
            run_check=run_check,
            black_config=black_config,
            what_for=" for other plugins",
        )
    if run_black_modules:
        paths = filter_paths(
            code_files,
            restrict=module_files,
            extensions=[".py"],
            with_cd=True,
        )
        _execute_black_for(
            session,
            paths=paths,
            run_check=run_check,
            black_config=black_modules_config or black_config,
            what_for=" for modules and module utils",
        )


def _execute_ruff_format_for(
    session: nox.Session,
    *,
    run_check: bool,
    files: list[Path],
    ruff_format_config: str | os.PathLike | None,
    what_for: str = "",
) -> None:
    if not files:
        session.warn(f"Skipping ruff format{what_for} (no files to process)")
        return
    command: list[str] = [
        "ruff",
        "format",
    ]
    if run_check:
        command.append("--check")
    if ruff_format_config is not None:
        command.extend(["--config", str(ruff_format_config)])
    command.extend(session.posargs)
    command.extend(str(file) for file in files)
    session.run(*command)


def _execute_ruff_format(
    session: nox.Session,
    *,
    run_check: bool,
    code_files: list[Path] | FileCollector,
    module_files: list[Path] | FileCollector,
    ruff_format_config: str | os.PathLike | None,
    ruff_format_modules_config: str | os.PathLike | None,
) -> None:
    files, files_modules, files_other = _get_files(
        code_files=code_files,
        module_files=module_files,
        split_modules=ruff_format_modules_config is not None
        and ruff_format_modules_config != ruff_format_config,
    )
    if files is not None:
        _execute_ruff_format_for(
            session,
            run_check=run_check,
            files=files,
            ruff_format_config=ruff_format_config,
        )
    if files_modules is not None:
        _execute_ruff_format_for(
            session,
            run_check=run_check,
            files=files_modules,
            ruff_format_config=ruff_format_modules_config or ruff_format_config,
            what_for=" for modules and module utils",
        )
    if files_other is not None:
        _execute_ruff_format_for(
            session,
            run_check=run_check,
            files=files_other,
            ruff_format_config=ruff_format_config,
            what_for=" for other files",
        )


def _execute_ruff_autofix_for(
    session: nox.Session,
    *,
    old_cwd: Path,
    root_dir: Path,
    collection_dir: Path,
    run_check: bool,
    files: list[Path],
    ruff_autofix_config: str | os.PathLike | None,
    ruff_autofix_select: list[str],
    what_for: str = "",
) -> None:
    if not files:
        session.warn(f"Skipping ruff autofix{what_for} (no files to process)")
        return
    command: list[str] = [
        "ruff",
        "check",
    ]
    if not run_check:
        command.append("--fix")
    if ruff_autofix_config is not None:
        command.extend(
            [
                "--config",
                str(
                    relative_to_walk_up(
                        (old_cwd / ruff_autofix_config).resolve(), root_dir
                    )
                ),
            ]
        )
    if ruff_autofix_select:
        command.extend(["--select", ",".join(ruff_autofix_select)])
    command.extend(session.posargs)
    relative_dir = collection_dir.relative_to(root_dir)
    for file in files:
        command.append(str(relative_dir / file))
    session.run(*command)


def _execute_ruff_autofix(
    session: nox.Session,
    *,
    root_dir: Path,
    collection_dir: Path,
    run_check: bool,
    code_files: list[Path] | FileCollector,
    module_files: list[Path] | FileCollector,
    ruff_autofix_config: str | os.PathLike | None,
    ruff_autofix_modules_config: str | os.PathLike | None,
    ruff_autofix_select: list[str],
) -> None:
    files, files_modules, files_other = _get_files(
        code_files=code_files,
        module_files=module_files,
        split_modules=ruff_autofix_modules_config is not None
        and ruff_autofix_modules_config != ruff_autofix_config,
    )
    old_cwd = Path.cwd()
    with session.chdir(root_dir):
        if files is not None:
            _execute_ruff_autofix_for(
                session,
                old_cwd=old_cwd,
                root_dir=root_dir,
                collection_dir=collection_dir,
                run_check=run_check,
                files=files,
                ruff_autofix_config=ruff_autofix_config,
                ruff_autofix_select=ruff_autofix_select,
            )
        if files_modules is not None:
            _execute_ruff_autofix_for(
                session,
                old_cwd=old_cwd,
                root_dir=root_dir,
                collection_dir=collection_dir,
                run_check=run_check,
                files=files_modules,
                ruff_autofix_config=ruff_autofix_modules_config or ruff_autofix_config,
                ruff_autofix_select=ruff_autofix_select,
                what_for=" for modules and module utils",
            )
        if files_other is not None:
            _execute_ruff_autofix_for(
                session,
                old_cwd=old_cwd,
                root_dir=root_dir,
                collection_dir=collection_dir,
                run_check=run_check,
                files=files_other,
                ruff_autofix_config=ruff_autofix_config,
                ruff_autofix_select=ruff_autofix_select,
                what_for=" for other files",
            )


def add_formatters(
    *,
    code_files: list[Path] | FileCollector,
    module_files: list[Path] | FileCollector,
    # isort:
    run_isort: bool,
    isort_config: str | os.PathLike | None,
    isort_modules_config: str | os.PathLike | None,
    isort_package: PackageTypeOrList,
    # black:
    run_black: bool,
    run_black_modules: bool | None,
    black_config: str | os.PathLike | None,
    black_modules_config: str | os.PathLike | None,
    black_package: PackageTypeOrList,
    # ruff format:
    run_ruff_format: bool,
    ruff_format_config: str | os.PathLike | None,
    ruff_format_modules_config: str | os.PathLike | None,
    ruff_format_package: PackageTypeOrList,
    # ruff autofix:
    run_ruff_autofix: bool,
    ruff_autofix_config: str | os.PathLike | None,
    ruff_autofix_modules_config: str | os.PathLike | None,
    ruff_autofix_package: PackageTypeOrList,
    ruff_autofix_select: list[str],
) -> None:
    """
    Add nox session for formatters.
    """
    if run_black_modules is None:
        run_black_modules = run_black
    run_check = IN_CI

    def compose_dependencies(session: nox.Session) -> list[PackageType]:
        deps = []
        if run_isort:
            deps.extend(
                check_package_types(
                    session,
                    "sessions.lint.isort_package",
                    normalize_package_type(isort_package),
                )
            )
        if run_black or run_black_modules:
            deps.extend(
                check_package_types(
                    session,
                    "sessions.lint.black_package",
                    normalize_package_type(black_package),
                )
            )
        if (
            run_ruff_format
            and run_ruff_autofix
            and ruff_format_package == ruff_autofix_package
        ):
            deps.extend(
                check_package_types(
                    session,
                    "sessions.lint.ruff_format_package",
                    normalize_package_type(ruff_format_package),
                )
            )
        else:
            if run_ruff_format:
                deps.extend(
                    check_package_types(
                        session,
                        "sessions.lint.ruff_format_package",
                        normalize_package_type(ruff_format_package),
                    )
                )
            if run_ruff_autofix:
                deps.extend(
                    check_package_types(
                        session,
                        "sessions.lint.ruff_autofix_package",
                        normalize_package_type(ruff_autofix_package),
                    )
                )
        return deps

    def formatters(session: nox.Session) -> None:
        install(session, *compose_dependencies(session))
        if run_isort or run_ruff_autofix:
            cwd = Path.cwd()
            cd = load_collection_data_from_disk(cwd)
            root_dir = Path(session.create_tmp()).resolve() / "collection-root"
            namespace_dir = root_dir / "ansible_collections" / cd.namespace
            namespace_dir.mkdir(parents=True, exist_ok=True)
            collection_path = namespace_dir / cd.name
            if not collection_path.exists():
                collection_path.symlink_to(
                    relative_to_walk_up(cwd, namespace_dir),
                    target_is_directory=True,
                )
        if run_isort:
            _execute_isort(
                session,
                root_dir=root_dir,
                collection_dir=collection_path,
                run_check=run_check,
                code_files=code_files,
                module_files=module_files,
                isort_config=isort_config,
                isort_modules_config=isort_modules_config,
            )
        if run_black or run_black_modules:
            _execute_black(
                session,
                run_check=run_check,
                code_files=code_files,
                module_files=module_files,
                run_black=run_black,
                run_black_modules=run_black_modules,
                black_config=black_config,
                black_modules_config=black_modules_config,
            )
        if run_ruff_format:
            _execute_ruff_format(
                session,
                run_check=run_check,
                code_files=code_files,
                module_files=module_files,
                ruff_format_config=ruff_format_config,
                ruff_format_modules_config=ruff_format_modules_config,
            )
        if run_ruff_autofix:
            _execute_ruff_autofix(
                session,
                root_dir=root_dir,
                collection_dir=collection_path,
                run_check=run_check,
                code_files=code_files,
                module_files=module_files,
                ruff_autofix_config=ruff_autofix_config,
                ruff_autofix_modules_config=ruff_autofix_modules_config,
                ruff_autofix_select=ruff_autofix_select,
            )

    formatters.__doc__ = compose_description(
        prefix={
            "one": "Run code formatter:",
            "other": "Run code formatters:",
        },
        programs={
            "isort": run_isort,
            "black": run_black,
            "ruff format": run_ruff_format,
            "ruff check --fix": run_ruff_autofix,
        },
    )
    nox.session(name="formatters", default=False)(formatters)


def add_codeqa(  # noqa: C901
    *,
    code_files: list[Path] | FileCollector,
    code_files_pylint: list[Path] | FileCollector,
    module_files: list[Path] | FileCollector,
    # ruff check:
    run_ruff_check: bool,
    ruff_check_config: str | os.PathLike | None,
    ruff_check_modules_config: str | os.PathLike | None,
    ruff_check_package: PackageTypeOrList,
    # flake8:
    run_flake8: bool,
    flake8_config: str | os.PathLike | None,
    flake8_modules_config: str | os.PathLike | None,
    flake8_package: PackageTypeOrList,
    # pylint:
    run_pylint: bool,
    pylint_rcfile: str | os.PathLike | None,
    pylint_modules_rcfile: str | os.PathLike | None,
    pylint_package: PackageTypeOrList,
    pylint_ansible_core_package: PackageTypeOrList | None,
    pylint_extra_deps: list[str | PackageType],
) -> None:
    """
    Add nox session for codeqa.
    """

    def compose_dependencies(session: nox.Session) -> list[PackageType]:
        deps = []
        if run_ruff_check:
            deps.extend(
                check_package_types(
                    session,
                    "sessions.lint.ruff_check_package",
                    normalize_package_type(ruff_check_package),
                )
            )
        if run_flake8:
            deps.extend(
                check_package_types(
                    session,
                    "sessions.lint.flake8_package",
                    normalize_package_type(flake8_package),
                )
            )
        if run_pylint:
            deps.extend(
                check_package_types(
                    session,
                    "sessions.lint.pylint_package",
                    normalize_package_type(pylint_package),
                )
            )
            deps.extend(
                check_package_types(
                    session,
                    "sessions.lint.pylint_ansible_core_package",
                    normalize_package_type(pylint_ansible_core_package),
                )
            )
            if os.path.isdir("tests/unit"):
                deps.append("pytest")
                if os.path.isfile("tests/unit/requirements.txt"):
                    deps.extend(["-r", "tests/unit/requirements.txt"])
            for idx, extra_dep in enumerate(pylint_extra_deps):
                deps.extend(
                    _split_arg(
                        session, extra_dep, "sessions.lint.pylint_extra_deps", idx
                    )
                )
        return deps

    def execute_ruff_check_impl(
        session: nox.Session,
        prepared_collections: CollectionSetup,
        *,
        files: list[Path],
        config: str | os.PathLike | None,
        what_for: str = "",
    ) -> list[Message]:
        if not files:
            session.warn(f"Skipping ruff check{what_for} (no files to process)")
            return []
        command: list[str] = [
            "ruff",
            "check",
            "--no-respect-gitignore",
            "--output-format=json",
        ]
        if config is not None:
            command.extend(
                [
                    "--config",
                    os.path.join(prepared_collections.current_collection.path, config),
                ]
            )
        command.extend(session.posargs)
        command.extend(
            str(path) for path in prepared_collections.prefix_current_paths(files)
        )
        # https://docs.astral.sh/ruff/linter/#exit-codes
        output = session.run(*command, silent=True, success_codes=[0, 1])

        return (
            parse_ruff_check_errors(
                source_path=prepared_collections.current_path,
                output=output,
            )
            if output
            else []
        )

    def execute_ruff_check(
        session: nox.Session,
        prepared_collections: CollectionSetup,
    ) -> None:
        files, files_modules, files_other = _get_files(
            code_files=code_files,
            module_files=module_files,
            split_modules=ruff_check_modules_config is not None
            and ruff_check_modules_config != ruff_check_config,
        )

        messages = []
        with session.chdir(prepared_collections.current_place), silence_run_verbosity():
            if files is not None:
                messages.extend(
                    execute_ruff_check_impl(
                        session,
                        prepared_collections,
                        files=files,
                        config=ruff_check_config,
                    )
                )
            if files_modules is not None:
                messages.extend(
                    execute_ruff_check_impl(
                        session,
                        prepared_collections,
                        files=files_modules,
                        config=ruff_check_modules_config or ruff_check_config,
                        what_for=" for modules and module utils",
                    )
                )
            if files_other is not None:
                messages.extend(
                    execute_ruff_check_impl(
                        session,
                        prepared_collections,
                        files=files_other,
                        config=ruff_check_config,
                        what_for=" for other files",
                    )
                )

        print_messages(
            session=session,
            messages=messages,
            fail_msg="Ruff check failed",
        )

    def execute_flake8_impl(
        session: nox.Session,
        *,
        files: list[Path],
        config: str | os.PathLike | None,
        what_for: str = "",
    ) -> None:
        if not files:
            session.warn(f"Skipping flake8{what_for} (no files to process)")
            return
        command: list[str] = [
            "flake8",
        ]
        if config is not None:
            command.extend(["--config", str(config)])
        command.extend(session.posargs)
        command.extend(str(file) for file in files)
        session.run(*command)

    def execute_flake8(session: nox.Session) -> None:
        files, files_modules, files_other = _get_files(
            code_files=code_files,
            module_files=module_files,
            split_modules=flake8_modules_config is not None
            and flake8_modules_config != flake8_config,
        )
        if files is not None:
            execute_flake8_impl(session, files=files, config=flake8_config)
        if files_modules is not None:
            execute_flake8_impl(
                session,
                files=files_modules,
                config=flake8_modules_config or flake8_config,
                what_for=" for modules and module utils",
            )
        if files_other is not None:
            execute_flake8_impl(
                session,
                files=files_other,
                config=flake8_config,
                what_for=" for other files",
            )

    def execute_pylint_impl(
        session: nox.Session,
        prepared_collections: CollectionSetup,
        config: os.PathLike | str | None,
        paths: list[Path],
        what_for: str = "",
    ) -> list[Message]:
        if not paths:
            session.warn(f"Skipping pylint{what_for} (no files to process)")
            return []
        command = ["pylint"]
        if config is not None:
            command.extend(
                [
                    "--rcfile",
                    os.path.join(prepared_collections.current_collection.path, config),
                ]
            )
        command.extend(["--source-roots", "."])
        command.extend(["--output-format", "json2"])
        command.extend(session.posargs)
        command.extend(
            str(path) for path in prepared_collections.prefix_current_paths(paths)
        )
        with silence_run_verbosity(), silence_run_verbosity():
            # Exit code is OR of some of 1, 2, 4, 8, 16
            output = session.run(
                *command, silent=True, success_codes=list(range(0, 32))
            )

        return (
            parse_pylint_json2_errors(
                source_path=prepared_collections.current_path,
                output=output,
            )
            if output
            else []
        )

    def execute_pylint(
        session: nox.Session, prepared_collections: CollectionSetup
    ) -> None:
        files, files_modules, files_other = _get_files(
            code_files=code_files_pylint,
            module_files=module_files,
            split_modules=pylint_modules_rcfile is not None
            and pylint_modules_rcfile != pylint_rcfile,
            cd_add_python_deps="importing-changed",
        )

        messages = []
        with session.chdir(prepared_collections.current_place):
            if files is not None:
                messages.extend(
                    execute_pylint_impl(
                        session,
                        prepared_collections,
                        pylint_modules_rcfile or pylint_rcfile,
                        files,
                    )
                )
            if files_modules is not None:
                messages.extend(
                    execute_pylint_impl(
                        session,
                        prepared_collections,
                        pylint_modules_rcfile or pylint_rcfile,
                        files_modules,
                        what_for=" for modules and module utils",
                    )
                )
            if files_other is not None:
                messages.extend(
                    execute_pylint_impl(
                        session,
                        prepared_collections,
                        pylint_modules_rcfile or pylint_rcfile,
                        files_other,
                        what_for=" for other files",
                    )
                )

        print_messages(session=session, messages=messages, fail_msg="Pylint failed")

    def codeqa(session: nox.Session) -> None:
        install(session, *compose_dependencies(session))
        prepared_collections: CollectionSetup | None = None
        if run_ruff_check or run_pylint:
            prepared_collections = prepare_collections(
                session,
                install_in_site_packages=False,
                extra_deps_files=["tests/unit/requirements.yml"],
            )
            if not prepared_collections:
                session.warn("Skipping pylint...")
        if run_ruff_check and prepared_collections:
            execute_ruff_check(session, prepared_collections)
        if run_flake8:
            execute_flake8(session)
        if run_pylint and prepared_collections:
            execute_pylint(session, prepared_collections)

    codeqa.__doc__ = compose_description(
        prefix={
            "other": "Run code QA:",
        },
        programs={
            "ruff check": run_ruff_check,
            "flake8": run_flake8,
            "pylint": run_pylint,
        },
    )
    nox.session(name="codeqa", default=False)(codeqa)


def add_yamllint(
    *,
    run_yamllint: bool,
    yamllint_config: str | os.PathLike | None,
    yamllint_config_plugins: str | os.PathLike | None,
    yamllint_config_plugins_examples: str | os.PathLike | None,
    yamllint_config_extra_docs: str | os.PathLike | None,
    yamllint_package: PackageTypeOrList,
    yamllint_antsibull_docutils_package: PackageTypeOrList,
) -> None:
    """
    Add yamllint session for linting YAML files and plugin/module docs.
    """

    def compose_dependencies(session: nox.Session) -> list[PackageType]:
        deps = []
        if run_yamllint:
            deps.extend(
                check_package_types(
                    session,
                    "sessions.lint.yamllint_package",
                    normalize_package_type(yamllint_package),
                )
            )
            deps.extend(
                check_package_types(
                    session,
                    "sessions.lint.yamllint_antsibull_docutils_package",
                    normalize_package_type(yamllint_antsibull_docutils_package),
                )
            )
        return deps

    def to_str(config: str | os.PathLike | None) -> str | None:
        return str(config) if config else None

    def execute_yamllint(session: nox.Session) -> None:
        all_files = list_all_files()
        all_yaml_filenames = filter_files_cd(
            [
                file
                for file in all_files
                if file.name.lower().endswith((".yml", ".yaml"))
            ],
        )
        if not all_yaml_filenames:
            session.warn("Skipping yamllint (no files to process)")
            return

        run_bare_script(
            session,
            "file-yamllint",
            use_session_python=True,
            files=all_yaml_filenames,
            extra_data={
                "config": to_str(yamllint_config),
            },
            process_messages=True,
        )

    def execute_plugin_yamllint(session: nox.Session) -> None:
        all_files = list_all_files()
        cwd = Path.cwd()
        plugins_dir = cwd / "plugins"
        ignore_dirs = [
            plugins_dir / "action",
            plugins_dir / "module_utils",
            plugins_dir / "plugin_utils",
        ]
        all_plugin_files = filter_files_cd(
            [
                file
                for file in all_files
                if file.is_relative_to(plugins_dir)
                and file.name.lower().endswith((".py", ".yml", ".yaml"))
                and not any(file.is_relative_to(dir) for dir in ignore_dirs)
            ],
        )
        if not all_plugin_files:
            session.warn("Skipping yamllint for modules/plugins (no files to process)")
            return
        run_bare_script(
            session,
            "plugin-yamllint",
            use_session_python=True,
            files=all_plugin_files,
            extra_data={
                "config": to_str(yamllint_config_plugins or yamllint_config),
                "config_examples": to_str(
                    yamllint_config_plugins_examples
                    or yamllint_config_plugins
                    or yamllint_config
                ),
            },
            process_messages=True,
        )

    def execute_extra_docs_yamllint(session: nox.Session) -> None:
        all_extra_docs = filter_files_cd(find_extra_docs_rst_files())
        if not all_extra_docs:
            session.warn("Skipping yamllint for extra docs (no files to process)")
            return
        run_bare_script(
            session,
            "rst-yamllint",
            use_session_python=True,
            files=all_extra_docs,
            extra_data={
                "config": to_str(
                    yamllint_config_extra_docs
                    or yamllint_config_plugins_examples
                    or yamllint_config_plugins
                    or yamllint_config
                ),
            },
            process_messages=True,
        )

    def yamllint(session: nox.Session) -> None:
        install(session, *compose_dependencies(session))
        if run_yamllint:
            execute_yamllint(session)
            execute_plugin_yamllint(session)
            execute_extra_docs_yamllint(session)

    yamllint.__doc__ = compose_description(
        prefix={
            "one": "Run YAML checker:",
            "other": "Run YAML checkers:",
        },
        programs={
            "yamllint": run_yamllint,
        },
    )
    nox.session(name="yamllint", default=False)(yamllint)


def add_typing(
    *,
    code_files: list[Path] | FileCollector,
    module_files: list[Path] | FileCollector,
    run_mypy: bool,
    mypy_config: str | os.PathLike | None,
    mypy_modules_config: str | os.PathLike | None,
    mypy_package: PackageTypeOrList,
    mypy_ansible_core_package: PackageTypeOrList | None,
    mypy_extra_deps: list[str | PackageType],
) -> None:
    """
    Add nox session for typing.
    """

    def compose_dependencies(session: nox.Session) -> list[PackageType]:
        deps = []
        if run_mypy:
            deps.extend(
                check_package_types(
                    session,
                    "sessions.lint.mypy_package",
                    normalize_package_type(mypy_package),
                )
            )
            if mypy_ansible_core_package is not None:
                deps.extend(
                    check_package_types(
                        session,
                        "sessions.lint.mypy_ansible_core_package",
                        normalize_package_type(mypy_ansible_core_package),
                    )
                )
            if os.path.isdir("tests/unit"):
                deps.append("pytest")
                if os.path.isfile("tests/unit/requirements.txt"):
                    deps.extend(["-r", "tests/unit/requirements.txt"])
            for idx, extra_dep in enumerate(mypy_extra_deps):
                deps.extend(
                    _split_arg(session, extra_dep, "sessions.lint.mypy_extra_deps", idx)
                )
        return deps

    def execute_mypy_impl(
        session: nox.Session,
        prepared_collections: CollectionSetup,
        *,
        files: list[Path],
        config: str | os.PathLike | None,
        what_for: str = "",
    ) -> list[Message]:
        files = prepared_collections.prefix_current_paths(files)
        if not files:
            session.warn(f"Skipping mypy{what_for} (no files to process)")
            return []
        command = ["mypy"]
        if config is not None:
            command.extend(
                [
                    "--config-file",
                    os.path.join(prepared_collections.current_collection.path, config),
                ]
            )
        command.append("--namespace-packages")
        command.append("--explicit-package-bases")
        command.extend(["--output", "json"])
        command.extend(session.posargs)
        command.extend(str(file) for file in files)
        with silence_run_verbosity():
            output = session.run(
                *command,
                env={"MYPYPATH": str(prepared_collections.current_place)},
                silent=True,
                success_codes=(0, 1, 2),
            )

        return (
            parse_mypy_errors(
                root_path=prepared_collections.current_place,
                source_path=prepared_collections.current_path,
                output=output,
            )
            if output
            else []
        )

    def execute_mypy(
        session: nox.Session, prepared_collections: CollectionSetup
    ) -> None:
        files, files_modules, files_other = _get_files(
            code_files=code_files,
            module_files=module_files,
            split_modules=mypy_modules_config is not None
            and mypy_modules_config != mypy_config,
            cd_add_python_deps="importing-changed",
        )

        messages = []
        with session.chdir(prepared_collections.current_place):
            if files is not None:
                messages.extend(
                    execute_mypy_impl(
                        session, prepared_collections, files=files, config=mypy_config
                    )
                )
            if files_modules is not None:
                messages.extend(
                    execute_mypy_impl(
                        session,
                        prepared_collections,
                        files=files_modules,
                        config=mypy_modules_config or mypy_config,
                        what_for=" for modules and module utils",
                    )
                )
            if files_other is not None:
                messages.extend(
                    execute_mypy_impl(
                        session,
                        prepared_collections,
                        files=files_other,
                        config=mypy_config,
                        what_for=" for other files",
                    )
                )

        print_messages(
            session=session,
            messages=messages,
            fail_msg="Mypy failed",
        )

    def typing(session: nox.Session) -> None:
        install(session, *compose_dependencies(session))
        prepared_collections = prepare_collections(
            session,
            install_in_site_packages=False,
            extra_deps_files=["tests/unit/requirements.yml"],
        )
        if not prepared_collections:
            session.warn("Skipping mypy...")
        if run_mypy and prepared_collections:
            execute_mypy(session, prepared_collections)

    typing.__doc__ = compose_description(
        prefix={
            "one": "Run type checker:",
            "other": "Run type checkers:",
        },
        programs={
            "mypy": run_mypy,
        },
    )
    nox.session(name="typing", default=False)(typing)


def add_config_lint(
    *,
    run_antsibullnox_config_lint: bool,
):
    """
    Add nox session for antsibull-nox config linting.
    """

    def antsibull_nox_config(session: nox.Session) -> None:
        if run_antsibullnox_config_lint:
            run_bare_script(
                session,
                "antsibull-nox-lint-config",
                process_messages=True,
            )

    antsibull_nox_config.__doc__ = "Lint antsibull-nox config"
    nox.session(name="antsibull-nox-config", python=False, default=False)(
        antsibull_nox_config
    )


def add_lint_sessions(
    *,
    make_lint_default: bool = True,
    code_files: list[Path] | FileCollector | None = None,
    extra_code_files: list[str] | None = None,
    module_files: list[Path] | FileCollector | None = None,
    # isort:
    run_isort: bool = True,
    isort_config: str | os.PathLike | None = None,
    isort_modules_config: str | os.PathLike | None = None,
    isort_package: PackageTypeOrList = "isort",
    # black:
    run_black: bool = True,
    run_black_modules: bool | None = None,
    black_config: str | os.PathLike | None = None,
    black_modules_config: str | os.PathLike | None = None,
    black_package: PackageTypeOrList = "black",
    # ruff format:
    run_ruff_format: bool = False,
    ruff_format_config: str | os.PathLike | None = None,
    ruff_format_modules_config: str | os.PathLike | None = None,
    ruff_format_package: PackageTypeOrList = "ruff",
    # ruff autofix:
    run_ruff_autofix: bool = False,
    ruff_autofix_config: str | os.PathLike | None = None,
    ruff_autofix_modules_config: str | os.PathLike | None = None,
    ruff_autofix_package: PackageTypeOrList = "ruff",
    ruff_autofix_select: list[str] | None = None,
    # ruff check:
    run_ruff_check: bool = False,
    ruff_check_config: str | os.PathLike | None = None,
    ruff_check_modules_config: str | os.PathLike | None = None,
    ruff_check_package: PackageTypeOrList = "ruff",
    # flake8:
    run_flake8: bool = True,
    flake8_config: str | os.PathLike | None = None,
    flake8_modules_config: str | os.PathLike | None = None,
    flake8_package: PackageTypeOrList = "flake8",
    # pylint:
    run_pylint: bool = True,
    pylint_rcfile: str | os.PathLike | None = None,
    pylint_modules_rcfile: str | os.PathLike | None = None,
    pylint_package: PackageTypeOrList = "pylint",
    pylint_ansible_core_package: PackageTypeOrList | None = "ansible-core",
    pylint_extra_deps: list[str | PackageType] | None = None,
    # yamllint:
    run_yamllint: bool = False,
    yamllint_config: str | os.PathLike | None = None,
    yamllint_config_plugins: str | os.PathLike | None = None,
    yamllint_config_plugins_examples: str | os.PathLike | None = None,
    yamllint_config_extra_docs: str | os.PathLike | None = None,
    yamllint_package: PackageTypeOrList = "yamllint",
    yamllint_antsibull_docutils_package: PackageTypeOrList = "antsibull-docutils",
    # mypy:
    run_mypy: bool = True,
    mypy_config: str | os.PathLike | None = None,
    mypy_modules_config: str | os.PathLike | None = None,
    mypy_package: PackageTypeOrList = "mypy",
    mypy_ansible_core_package: PackageTypeOrList | None = "ansible-core",
    mypy_extra_deps: list[str | PackageType] | None = None,
    # antsibull-nox config lint:
    run_antsibullnox_config_lint: bool = True,
) -> None:
    """
    Add nox sessions for linting.
    """
    if code_files is not None:
        if extra_code_files is not None:
            raise ValueError("Cannot specify both code_files and extra_code_files")
        code_files_w_noxfile = code_files
    else:
        extra_code_files_paths = (
            [Path(file) for file in extra_code_files] if extra_code_files else []
        )
        code_files_w_noxfile = CODE_FILES_W_NOXFILE + extra_code_files_paths
        code_files = CODE_FILES + extra_code_files_paths

    if module_files is None:
        module_files = MODULE_PATHS

    has_formatters = (
        run_isort
        or run_black
        or run_black_modules
        or False
        or run_ruff_format
        or run_ruff_autofix
    )
    has_codeqa = run_ruff_check or run_flake8 or run_pylint
    has_yamllint = run_yamllint
    has_typing = run_mypy
    has_config_lint = run_antsibullnox_config_lint

    add_lint(
        has_formatters=has_formatters,
        has_codeqa=has_codeqa,
        has_yamllint=has_yamllint,
        has_typing=has_typing,
        has_config_lint=has_config_lint,
        make_lint_default=make_lint_default,
    )

    if has_formatters:
        add_formatters(
            code_files=code_files_w_noxfile,
            module_files=module_files,
            run_isort=run_isort,
            isort_config=isort_config,
            isort_modules_config=isort_modules_config,
            isort_package=isort_package,
            run_black=run_black,
            run_black_modules=run_black_modules,
            black_config=black_config,
            black_modules_config=black_modules_config,
            black_package=black_package,
            run_ruff_format=run_ruff_format,
            ruff_format_config=ruff_format_config,
            ruff_format_modules_config=ruff_format_modules_config,
            ruff_format_package=ruff_format_package,
            run_ruff_autofix=run_ruff_autofix,
            ruff_autofix_config=ruff_autofix_config,
            ruff_autofix_modules_config=ruff_autofix_modules_config,
            ruff_autofix_package=ruff_autofix_package,
            ruff_autofix_select=ruff_autofix_select or [],
        )

    if has_codeqa:
        add_codeqa(
            code_files=code_files_w_noxfile,
            code_files_pylint=code_files,
            module_files=module_files,
            run_ruff_check=run_ruff_check,
            ruff_check_config=ruff_check_config,
            ruff_check_modules_config=ruff_check_modules_config,
            ruff_check_package=ruff_check_package,
            run_flake8=run_flake8,
            flake8_config=flake8_config,
            flake8_modules_config=flake8_modules_config,
            flake8_package=flake8_package,
            run_pylint=run_pylint,
            pylint_rcfile=pylint_rcfile,
            pylint_modules_rcfile=pylint_modules_rcfile,
            pylint_package=pylint_package,
            pylint_ansible_core_package=pylint_ansible_core_package,
            pylint_extra_deps=pylint_extra_deps or [],
        )

    if has_yamllint:
        add_yamllint(
            run_yamllint=run_yamllint,
            yamllint_config=yamllint_config,
            yamllint_config_plugins=yamllint_config_plugins,
            yamllint_config_plugins_examples=yamllint_config_plugins_examples,
            yamllint_config_extra_docs=yamllint_config_extra_docs,
            yamllint_package=yamllint_package,
            yamllint_antsibull_docutils_package=yamllint_antsibull_docutils_package,
        )

    if has_typing:
        add_typing(
            code_files=code_files,
            module_files=module_files,
            run_mypy=run_mypy,
            mypy_config=mypy_config,
            mypy_modules_config=mypy_modules_config,
            mypy_package=mypy_package,
            mypy_ansible_core_package=mypy_ansible_core_package,
            mypy_extra_deps=mypy_extra_deps or [],
        )

    if has_config_lint:
        add_config_lint(
            run_antsibullnox_config_lint=run_antsibullnox_config_lint,
        )


__all__ = [
    "add_lint_sessions",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
VCS provider factory.
"""

from __future__ import annotations

import os
from pathlib import Path

from . import VCS, VcsProvider
from .git import GitProvider

GIT_EXECUTABLE_ENV_VAR = "ANTSIBULL_GIT_EXECUTABLE"


def get_vcs_provider(vcs: VCS) -> VcsProvider:
    """
    Factory method for VcsProvider instances.
    """
    if vcs == "git":
        git_executable = os.environ.get(GIT_EXECUTABLE_ENV_VAR) or "git"
        return GitProvider(git_executable=Path(git_executable))

    raise RuntimeError(f"Internal error: unhandled VCS {vcs}")


__all__ = ("get_vcs_provider",)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Python utilities for import relations.
"""

from __future__ import annotations

import ast
import dataclasses
import typing as t
from pathlib import Path

from ..paths.utils import path_walk

if t.TYPE_CHECKING:
    from collections.abc import Sequence


@dataclasses.dataclass(eq=True, frozen=True)
class PythonModule:
    """
    A Python module.
    """

    path: Path
    module_path: tuple[str, ...]
    imports: frozenset[PythonImport]
    parsable: bool


@dataclasses.dataclass(eq=True, frozen=True)
class PythonImport:
    """
    A symbol import in a Python module.
    """

    symbol_path: tuple[str, ...]

    def belongs_to(self, module_path: tuple[str, ...]) -> bool:
        """
        Determine whether the imported symbol is part of the module.
        """
        return module_path == self.symbol_path[: len(module_path)]


class ImportFinder(ast.NodeVisitor):
    """Visitor for finding all imports."""

    def __init__(self, module_path: tuple[str, ...]) -> None:
        """
        Create import finder for a module of given path.
        """
        self.module_path = list(module_path)
        self.imports: set[PythonImport] = set()
        self.imports_ok = True

    def visit_Import(self, node: ast.Import) -> None:  # pylint: disable=invalid-name
        """Visit an 'import' node."""
        self.generic_visit(node)
        for alias in node.names:
            self.imports.add(PythonImport(tuple(alias.name.split("."))))

    def visit_ImportFrom(  # pylint: disable=invalid-name
        self, node: ast.ImportFrom
    ) -> None:
        """Visit an 'import from' node."""
        self.generic_visit(node)
        base = node.module.split(".") if node.module else []
        if node.level > 0:
            if node.level > len(self.module_path):
                self.imports_ok = False
                return
            base = self.module_path[: -node.level] + base
        base.append("")
        for alias in node.names:
            base[-1] = alias.name
            self.imports.add(PythonImport(tuple(base)))


def get_imported_packages(
    path: Path, module_path: tuple[str, ...]
) -> tuple[set[PythonImport], bool]:
    """
    Load all Python imports from a Python file.
    """
    with path.open("rb") as f:
        contents = f.read()

    try:
        tree = ast.parse(contents)
    except SyntaxError:
        return set(), False

    finder = ImportFinder(module_path)
    finder.visit(tree)
    return finder.imports, finder.imports_ok


def get_module_data_from_module_path(
    path: Path, *, module_path: tuple[str, ...]
) -> PythonModule:
    """
    Given a Python file and a source root, load the module information.
    """
    imports, parsable = get_imported_packages(path, module_path)
    return PythonModule(
        path=path,
        module_path=module_path,
        imports=frozenset(imports),
        parsable=parsable,
    )


def get_module_path(relative_path: Path) -> tuple[str, ...]:
    """
    Given a relative path to a Python file form the source root, extract the module's path.
    """
    module_path = tuple(
        parent.name for parent in reversed(tuple(relative_path.parents)[:-1])
    )
    if relative_path.name == "__init__.py":
        return module_path
    return module_path + (relative_path.name.removesuffix(".py"),)


def get_module_data_from_source_root(path: Path, *, source_root: Path) -> PythonModule:
    """
    Given a Python file and a source root, load the module information.
    """
    relative_path = path.resolve().relative_to(source_root.resolve())
    module_path = get_module_path(relative_path)
    return get_module_data_from_module_path(path, module_path=module_path)


def get_all_module_data(
    paths_with_module_paths: Sequence[tuple[Path, tuple[str, ...]]],
) -> t.Generator[PythonModule]:
    """
    Given a source root and a sequence of paths in it,
    load information on all Python files in these paths.
    """
    for path, module_path in paths_with_module_paths:
        resolved_path = path.resolve()
        if resolved_path.is_dir():
            for dirpath, _, filenames in path_walk(path):
                reldir = dirpath.relative_to(path)
                for filename in filenames:
                    if not filename.endswith(".py"):
                        continue
                    rel_module_path = get_module_path(reldir / filename)
                    file_path = dirpath / filename
                    yield get_module_data_from_module_path(
                        file_path, module_path=module_path + rel_module_path
                    )
        elif resolved_path.is_file():
            yield get_module_data_from_module_path(path, module_path=module_path)


__all__ = (
    "get_all_module_data",
    "get_imported_packages",
    "get_module_data_from_module_path",
    "get_module_data_from_source_root",
    "get_module_path",
)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Python file dependency data.
"""

from __future__ import annotations

import functools
import typing as t
from dataclasses import dataclass
from pathlib import Path

from ..collection import load_collection_data_from_disk
from ..paths.utils import list_all_files
from .imports import PythonModule, get_module_data_from_module_path, get_module_path


@dataclass
class PythonDependencyInfo:
    """
    Stores Python dependency information.
    """

    file_to_module_path: dict[Path, tuple[str, ...]]
    module_path_to_file: dict[tuple[str, ...], Path]
    file_to_imported_modules: dict[
        Path, tuple[frozenset[tuple[str, ...]], frozenset[Path]]
    ]
    file_to_imported_by_modules: dict[
        Path, tuple[frozenset[tuple[str, ...]], frozenset[Path]]
    ]


@dataclass
class _Node:
    python_module: PythonModule | None
    children: dict[str, _Node]

    def find(
        self, module_path: tuple[str, ...]
    ) -> tuple[PythonModule, tuple[str, ...]] | None:
        """
        Find Python module and remaining path for module path.
        """
        next_node = self.children.get(module_path[0]) if len(module_path) > 0 else None
        if next_node is not None:
            return next_node.find(module_path[1:])
        if self.python_module is None:
            return None
        return self.python_module, module_path

    def add(
        self,
        python_module: PythonModule,
        *,
        relative_path: tuple[str, ...] | None = None,
    ) -> None:
        """
        Add Python module to tree.
        """
        if relative_path is None:
            relative_path = python_module.module_path
        if len(relative_path) == 0:
            if self.python_module is not None:
                raise ValueError(
                    f"Found multiple modules for {relative_path}:"
                    f" {python_module.path} vs. {self.python_module.path}"
                )
            self.python_module = python_module
            return
        next_node = self.children.get(relative_path[0])
        if next_node is None:
            next_node = _Node(None, {})
            self.children[relative_path[0]] = next_node
        next_node.add(python_module, relative_path=relative_path[1:])


def get_all_collection_module_data() -> t.Generator[PythonModule]:
    """
    Load all Python file module infos for the current collection.
    """
    cwd = Path.cwd()
    cd = load_collection_data_from_disk(cwd)
    module_root = ("ansible_collections", cd.namespace, cd.name)
    for path in list_all_files():
        if not path.name.endswith(".py"):
            continue
        module_path = module_root + get_module_path(path.relative_to(cwd))
        yield get_module_data_from_module_path(path, module_path=module_path)


def _get_collection_module_tree() -> tuple[list[PythonModule], _Node]:
    all_modules: list[PythonModule] = []
    root = _Node(None, {})
    for python_module in get_all_collection_module_data():
        all_modules.append(python_module)
        root.add(python_module)
    return all_modules, root


@functools.cache
def get_python_dependency_info() -> PythonDependencyInfo:
    """
    Extract Python dependency information for current collection.
    """
    all_modules, root = _get_collection_module_tree()
    file_to_module_path: dict[Path, tuple[str, ...]] = {}
    module_path_to_file: dict[tuple[str, ...], Path] = {}
    file_to_imported_modules: dict[
        Path, tuple[frozenset[tuple[str, ...]], frozenset[Path]]
    ] = {}
    file_to_imported_by_modules: dict[Path, tuple[set[tuple[str, ...]], set[Path]]] = {}
    for python_module in all_modules:
        file_to_module_path[python_module.path] = python_module.module_path
        module_path_to_file[python_module.module_path] = python_module.path
        file_to_imported_by_modules[python_module.path] = (set(), set())
    for python_module in all_modules:
        module_imports: set[tuple[str, ...]] = set()
        module_import_files: set[Path] = set()
        for python_import in python_module.imports:
            res = root.find(python_import.symbol_path)
            if res is not None:
                module_imports.add(res[0].module_path)
                module_import_files.add(res[0].path)
                imp_by_mp, imp_by_fs = file_to_imported_by_modules[res[0].path]
                imp_by_mp.add(python_module.module_path)
                imp_by_fs.add(python_module.path)
        file_to_imported_modules[python_module.path] = (
            frozenset(module_imports),
            frozenset(module_import_files),
        )
    return PythonDependencyInfo(
        file_to_module_path=file_to_module_path,
        module_path_to_file=module_path_to_file,
        file_to_imported_modules=file_to_imported_modules,
        file_to_imported_by_modules={
            k: (frozenset(v1), frozenset(v2))
            for k, (v1, v2) in file_to_imported_by_modules.items()
        },
    )


__all__ = (
    "PythonDependencyInfo",
    "get_python_dependency_info",
)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

from __future__ import annotations

from pathlib import Path
from unittest.mock import patch

from antsibull_nox.python.python_dependencies import (
    get_python_dependency_info,
)

from ..utils import chdir


def test_get_python_dependency_info(tmp_path: Path) -> None:
    galaxy_yml = tmp_path / "galaxy.yml"
    galaxy_yml.write_text(
        r"""
namespace: foo
name: bar
"""
    )

    plugins = tmp_path / "plugins"
    plugins.mkdir()
    module_utils = plugins / "module_utils"
    module_utils.mkdir()
    modules = plugins / "modules"
    modules.mkdir()

    module_utils_foo = module_utils / "foo.py"
    module_utils_foo.write_text(
        r"""
"""
    )

    module_utils_bar = module_utils / "bar"
    module_utils_bar.mkdir()

    module_utils_bar_init = module_utils_bar / "__init__.py"
    module_utils_bar_init.write_text(
        r"""
"""
    )

    module_utils_bar_foobarbaz = module_utils_bar / "foobarbaz.py"
    module_utils_bar_foobarbaz.write_text(
        r"""
"""
    )

    module_utils_foo = module_utils / "foo.py"
    module_utils_foo.write_text(
        r"""
"""
    )

    module_1 = modules / "module_1.py"
    module_1.write_text(
        r"""
from ansible.module_utils.basic import AnsibleModule
from ..module_utils import foo
from ansible_collections.foo.bar.plugins.module_utils.bar import run_bar, foobarbaz
from ansible_collections.foo.bar.plugins.module_utils.baz import run_baz
from ansible_collections.foo.bam.plugins.module_utils.bazbam import run_bazbam
"""
    )

    with chdir(tmp_path):
        with patch(
            "antsibull_nox.python.python_dependencies.list_all_files",
            return_value=[
                galaxy_yml,
                module_utils_foo,
                module_utils_bar_init,
                module_utils_bar_foobarbaz,
                module_1,
            ],
        ):
            get_python_dependency_info.cache_clear()
            result = get_python_dependency_info()

    assert result.file_to_module_path == {
        module_utils_foo: (
            "ansible_collections",
            "foo",
            "bar",
            "plugins",
            "module_utils",
            "foo",
        ),
        module_utils_bar_init: (
            "ansible_collections",
            "foo",
            "bar",
            "plugins",
            "module_utils",
            "bar",
        ),
        module_utils_bar_foobarbaz: (
            "ansible_collections",
            "foo",
            "bar",
            "plugins",
            "module_utils",
            "bar",
            "foobarbaz",
        ),
        module_1: (
            "ansible_collections",
            "foo",
            "bar",
            "plugins",
            "modules",
            "module_1",
        ),
    }
    assert result.module_path_to_file == {
        (
            "ansible_collections",
            "foo",
            "bar",
            "plugins",
            "module_utils",
            "foo",
        ): module_utils_foo,
        (
            "ansible_collections",
            "foo",
            "bar",
            "plugins",
            "module_utils",
            "bar",
        ): module_utils_bar_init,
        (
            "ansible_collections",
            "foo",
            "bar",
            "plugins",
            "module_utils",
            "bar",
            "foobarbaz",
        ): module_utils_bar_foobarbaz,
        (
            "ansible_collections",
            "foo",
            "bar",
            "plugins",
            "modules",
            "module_1",
        ): module_1,
    }
    assert result.file_to_imported_modules == {
        module_utils_foo: (frozenset(), frozenset()),
        module_utils_bar_init: (frozenset(), frozenset()),
        module_utils_bar_foobarbaz: (frozenset(), frozenset()),
        module_1: (
            frozenset(
                [
                    (
                        "ansible_collections",
                        "foo",
                        "bar",
                        "plugins",
                        "module_utils",
                        "foo",
                    ),
                    (
                        "ansible_collections",
                        "foo",
                        "bar",
                        "plugins",
                        "module_utils",
                        "bar",
                    ),
                    (
                        "ansible_collections",
                        "foo",
                        "bar",
                        "plugins",
                        "module_utils",
                        "bar",
                        "foobarbaz",
                    ),
                ]
            ),
            frozenset(
                [module_utils_foo, module_utils_bar_init, module_utils_bar_foobarbaz]
            ),
        ),
    }
    assert result.file_to_imported_by_modules == {
        module_utils_foo: (
            frozenset(
                [
                    (
                        "ansible_collections",
                        "foo",
                        "bar",
                        "plugins",
                        "modules",
                        "module_1",
                    )
                ]
            ),
            frozenset([module_1]),
        ),
        module_utils_bar_init: (
            frozenset(
                [
                    (
                        "ansible_collections",
                        "foo",
                        "bar",
                        "plugins",
                        "modules",
                        "module_1",
                    )
                ]
            ),
            frozenset([module_1]),
        ),
        module_utils_bar_foobarbaz: (
            frozenset(
                [
                    (
                        "ansible_collections",
                        "foo",
                        "bar",
                        "plugins",
                        "modules",
                        "module_1",
                    )
                ]
            ),
            frozenset([module_1]),
        ),
        module_1: (
            frozenset([]),
            frozenset([]),
        ),
    }



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Error reporting.
"""

from __future__ import annotations

import dataclasses
import enum


class Level(enum.Enum):
    """
    Message level.
    """

    INFO = 1
    WARNING = 2
    ERROR = 3


@dataclasses.dataclass(frozen=True)
class Location:
    """
    A location in a source file.
    """

    line: int
    column: int | None = None
    exact: bool = True

    def __get_tuple(self) -> tuple[int, bool, int, bool]:
        """Helper for comparison functions."""
        return self.line, self.column is not None, self.column or 0, self.exact

    def __lt__(self, other: Location) -> bool:
        o = other.__get_tuple()  # pylint: disable=protected-access
        return self.__get_tuple() < o

    def __le__(self, other: Location) -> bool:
        o = other.__get_tuple()  # pylint: disable=protected-access
        return self.__get_tuple() <= o

    def __gt__(self, other: Location) -> bool:
        o = other.__get_tuple()  # pylint: disable=protected-access
        return self.__get_tuple() > o

    def __ge__(self, other: Location) -> bool:
        o = other.__get_tuple()  # pylint: disable=protected-access
        return self.__get_tuple() >= o


@dataclasses.dataclass(frozen=True)
class Message:
    """
    A linter output message.
    """

    file: str | None
    # position and end_position point to the first and last affected
    # character, respectively. The special column -1 can be used for
    # end_position to indicate the position after the last character
    # in the line.
    position: Location | None
    end_position: Location | None
    level: Level
    id: str | None
    message: str
    symbol: str | None = None
    hint: str | None = None
    note: str | None = None
    url: str | None = None

    def __get_tuple(
        self,
    ) -> tuple[
        bool,
        str,
        bool,
        Location | None,
        bool,
        Location | None,
        Level,
        bool,
        str,
        str,
        bool,
        str,
        bool,
        str,
        bool,
        str,
        bool,
        str,
    ]:
        """Helper for comparison functions."""
        return (
            self.file is not None,
            self.file or "",
            self.position is not None,
            self.position,
            self.end_position is not None,
            self.end_position,
            self.level,
            self.id is not None,
            self.id or "",
            self.message,
            self.symbol is not None,
            self.symbol or "",
            self.hint is not None,
            self.hint or "",
            self.note is not None,
            self.note or "",
            self.url is not None,
            self.url or "",
        )

    def __lt__(self, other: Message) -> bool:
        o = other.__get_tuple()  # pylint: disable=protected-access
        return self.__get_tuple() < o

    def __le__(self, other: Message) -> bool:
        o = other.__get_tuple()  # pylint: disable=protected-access
        return self.__get_tuple() <= o

    def __gt__(self, other: Message) -> bool:
        o = other.__get_tuple()  # pylint: disable=protected-access
        return self.__get_tuple() > o

    def __ge__(self, other: Message) -> bool:
        o = other.__get_tuple()  # pylint: disable=protected-access
        return self.__get_tuple() >= o


__all__ = ("Level", "Location", "Message")



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Utils for handling Ansible values.
"""

from __future__ import annotations

import dataclasses
import os
import typing as t


@dataclasses.dataclass
class AnsibleValueExplicit:
    """
    An explicit value.
    """

    value: t.Any

    template_value: str | None = None

    def get_value(self) -> tuple[t.Any, bool]:
        """
        Get the current value, together with a boolean whether this value should be set.
        """
        return self.value, True


@dataclasses.dataclass
class AnsibleValueFromEnv:
    """
    A value taken from an environment variable
    """

    name: str
    fallback: t.Any = None
    unset_if_not_set: bool = False

    template_value: str | None = None

    def get_value(self) -> tuple[t.Any, bool]:
        """
        Get the current value, together with a boolean whether this value should be set.
        """
        if self.name in os.environ:
            return os.environ[self.name], True
        if self.unset_if_not_set:
            return None, False
        return self.fallback, True


AnsibleValue = t.Union[
    AnsibleValueExplicit,
    AnsibleValueFromEnv,
]


__all__ = [
    "AnsibleValueExplicit",
    "AnsibleValueFromEnv",
    "AnsibleValue",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Error reporting.
"""

from __future__ import annotations

_JSON_END: dict[str, str] = {
    "{": "}",
    "[": "]",
}
_JSON_START: tuple[str, ...] = tuple(_JSON_END.keys())


def find_json_line(output: str) -> str:
    """
    Given output of a program, find a JSON object or list in a single line.

    This function assumes that the line containing the JSON only contains
    whitespace before and after the JSON.
    """
    for line in output.splitlines():
        line = line.strip()
        if line.startswith(_JSON_START) and line.endswith(_JSON_END[line[0]]):
            return line
    return output


def find_json(output: str) -> str:
    """
    Given output of a program, find a JSON object or list in it.

    This function assumes that the object starts and ends in a line that does not
    have any noise preceeding / succeeding it.
    """
    lines = output.splitlines()
    for start, line in enumerate(lines):
        line = line.strip()
        if line.startswith(_JSON_START):
            end_char = _JSON_END[line[0]]
            break
    else:
        # Didn't find start
        return output
    lines = lines[start:]
    for end in range(len(lines) - 1, -1, -1):
        if lines[end].strip().endswith(end_char):
            break
    else:
        # Didn't find end
        return output
    lines = lines[: end + 1]
    return "\n".join(lines)


__all__ = ("find_json_line", "find_json")



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Generate job matrix for use in CI systems.
"""

from __future__ import annotations

import argparse
import json
import os
import sys
import typing as t
from collections.abc import Sequence

import nox

from ..ansible import get_actual_ansible_core_version, parse_ansible_core_version
from ..utils import Version
from .utils import get_registered_sessions, parse_args


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="nox -e matrix-generator --",
        description="Generate matrix for CI systems.",
        exit_on_error=False,
    )

    parser.add_argument("--min-ansible-core", help="Minimum ansible-core version")
    parser.add_argument("--max-ansible-core", help="Maximum ansible-core version")
    parser.add_argument(
        "--include-tags", help="Comma-separated list of tags that have to be present"
    )
    parser.add_argument(
        "--exclude-tags", help="Comma-separated list of tags that must not be present"
    )

    return parser


def _parse_version(
    version: str | None, *, option_name: str, session: nox.Session
) -> Version | None:
    if version is None:
        return None
    try:
        return get_actual_ansible_core_version(parse_ansible_core_version(version))
    except ValueError as exc:
        return session.error(f"{option_name}: {exc}")


def _ensure_sequence(value: str | Sequence[str] | None) -> Sequence[str]:
    if isinstance(value, str):
        return [value]
    if isinstance(value, Sequence):
        return value
    return []


def _filter(
    sessions: list[dict[str, t.Any]],
    *,
    min_ansible_core: Version | None,
    max_ansible_core: Version | None,
    include_tags: list[str] | None,
    exclude_tags: list[str] | None,
) -> list[dict[str, t.Any]]:
    result = []
    for session in sessions:
        ansible_core = session.get("ansible-core")
        if isinstance(ansible_core, str):
            version = get_actual_ansible_core_version(
                parse_ansible_core_version(ansible_core)
            )
            if min_ansible_core is not None and version < min_ansible_core:
                continue
            if max_ansible_core is not None and version > max_ansible_core:
                continue
        tags = _ensure_sequence(session.get("tags"))
        if include_tags and not all(tag in tags for tag in include_tags):
            continue
        if exclude_tags and any(tag in tags for tag in exclude_tags):
            continue
        result.append(session)
    return result


def _split(tag_string: str | None) -> t.Generator[str]:
    if tag_string is None:
        return
    for part in tag_string.split(","):
        part = part.strip()
        if part:
            yield part


def add_matrix_generator() -> None:
    """
    Add a session that generates matrixes for CI systems.
    """

    def matrix_generator(
        session: nox.Session,
    ) -> None:
        parser = _create_parser()
        args = parse_args(parser=parser, session=session)
        if args is None:
            return

        min_ansible_core = _parse_version(
            args.min_ansible_core, option_name="--min-ansible-core", session=session
        )
        max_ansible_core = _parse_version(
            args.max_ansible_core, option_name="--max-ansible-core", session=session
        )
        include_tags = list(_split(args.include_tags))
        exclude_tags = list(_split(args.exclude_tags))

        registered_sessions = get_registered_sessions()
        for key, sessions in list(registered_sessions.items()):
            filtered_sessions = _filter(
                sessions,
                min_ansible_core=min_ansible_core,
                max_ansible_core=max_ansible_core,
                include_tags=include_tags,
                exclude_tags=exclude_tags,
            )
            if filtered_sessions:
                registered_sessions[key] = filtered_sessions
            else:
                del registered_sessions[key]

        json_output = os.environ.get("ANTSIBULL_NOX_MATRIX_JSON")
        if json_output:
            print(f"Writing JSON output to {json_output}...")
            with open(json_output, "wt", encoding="utf-8") as f:
                f.write(json.dumps(registered_sessions))

        github_output = os.environ.get("GITHUB_OUTPUT")
        if github_output:
            print(f"Writing GitHub output to {github_output}...")
            with open(github_output, "at", encoding="utf-8") as f:
                for name, sessions in registered_sessions.items():
                    f.write(f"{name}={json.dumps(sessions)}\n")

        for name, sessions in sorted(registered_sessions.items()):
            print(f"{name} ({len(sessions)}):")
            for session_data in sessions:
                data = session_data.copy()
                session_name = data.pop("name")
                print(f"  {session_name}: {data}")

        sys.stdout.flush()

    matrix_generator.__doc__ = "Generate matrix for CI systems."
    nox.session(
        name="matrix-generator",
        python=False,
        default=False,
    )(matrix_generator)


__all__ = [
    "add_matrix_generator",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Utils for dealing with packages.
"""

from __future__ import annotations

import json

import nox
from packaging.version import InvalidVersion
from packaging.version import parse as parse_version

from .scripts import run_bare_script


def get_package_versions(
    session: nox.Session,
    /,
    packages: list[str] | str,
    *,
    use_session_python: bool = True,
) -> None | dict[str, str | None]:
    """
    Retrieve the versions of one or more Python packages.
    """
    name = "get-package-versions"
    if isinstance(packages, str):
        packages = [packages]
    if not packages:
        return {}
    result = run_bare_script(
        session,
        name,
        use_session_python=use_session_python,
        files=[],
        extra_data={"packages": packages},
        silent=True,
    )
    if result is None:
        return None
    return json.loads(result)


def get_package_version(
    session: nox.Session,
    /,
    package: str,
    *,
    use_session_python: bool = True,
) -> str | None:
    """
    Retrieve a Python package's version.
    """
    result = get_package_versions(
        session, package, use_session_python=use_session_python
    )
    return None if result is None else result.get(package)


def is_new_enough(actual_version: str | None, *, min_version: str) -> bool:
    """
    Given a program version, compares it to the min_version.
    If the program version is not given, it is assumed to be "new enough".
    """
    if actual_version is None:
        return True
    try:
        act_v = parse_version(actual_version)
    except InvalidVersion as exc:
        raise ValueError(
            f"Cannot parse actual version {actual_version!r}: {exc}"
        ) from exc
    try:
        min_v = parse_version(min_version)
    except InvalidVersion as exc:
        raise ValueError(
            f"Cannot parse minimum version {min_version!r}: {exc}"
        ) from exc
    return act_v >= min_v


__all__ = [
    "get_package_version",
    "get_package_versions",
    "is_new_enough",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Git specific code.
"""

from __future__ import annotations

import subprocess
import typing as t
from pathlib import Path

from . import VcsProvider


class GitProvider(VcsProvider):
    """
    Provides an interface to Git.
    """

    def __init__(self, git_executable: Path) -> None:
        """
        Create a GitProvider instance.
        """
        self.git_executable = git_executable

    @t.overload
    def run(
        self,
        *args: str | Path,
        check_rc: bool = False,
        decode: t.Literal[False] = False,
    ) -> tuple[int, bytes, str]: ...

    @t.overload
    def run(
        self, *args: str | Path, check_rc: bool = False, decode: t.Literal[True]
    ) -> tuple[int, str, str]: ...

    def run(
        self, *args: str | Path, check_rc: bool = False, decode: bool = False
    ) -> tuple[int, str | bytes, str]:
        """
        Run a git command.
        """
        command = [self.git_executable] + [str(arg) for arg in args]
        p = subprocess.run(command, capture_output=True, check=check_rc)
        return (
            p.returncode,
            p.stdout.decode("utf-8") if decode else p.stdout,
            p.stderr.decode("utf-8"),
        )

    def get_all_local_branches(self, *, repo: Path) -> list[str]:
        """
        Get a list of all local branches of the repo.
        """
        _, stdout, __ = self.run(
            "-C",
            repo,
            "for-each-ref",
            "--format",
            "%(refname:strip=2)",
            "--",
            "refs/heads/",
            check_rc=True,
            decode=True,
        )
        return [entry for entry in stdout.splitlines() if entry]

    def get_all_matching_local_branches(
        self, *, repo: Path, branch_patterns: list[str]
    ) -> list[str]:
        """
        Get a list of all local branches of the repo that match one of the given patterns.
        """
        if not branch_patterns:
            return []
        _, stdout, __ = self.run(
            "-C",
            repo,
            "for-each-ref",
            "--format",
            "%(refname:strip=2)",
            "--",
            *(f"refs/heads/{pattern}" for pattern in branch_patterns),
            check_rc=True,
            decode=True,
        )
        return [entry for entry in stdout.splitlines() if entry]

    def get_list(
        self,
        *args: str | Path,
        repo: Path,
        separator: str,
    ) -> list[str]:
        """
        Execute a command that returns a list of strings separated by ``separator``.
        """
        _, stdout, __ = self.run(
            "-C",
            repo,
            *args,
            check_rc=True,
            decode=True,
        )
        to_split = stdout.removesuffix(separator)
        if not to_split:
            return []
        return list(to_split.split(separator))

    def get_branch_fork_point(self, *, repo: Path, branch: str) -> str | None:
        """
        Return a reference to the point at which the given branch was forked.
        """
        # See https://github.com/ansible/ansible/pull/79734 for details.
        rc, stdout, _ = self.run(
            "-C", repo, "merge-base", "--", branch, "HEAD", decode=True
        )
        out = stdout.strip()
        if rc != 0 or not out:
            return None
        return out

    def find_repo_path(self, *, path: Path) -> Path | None:
        """
        Given a path, finds the associated repository path for this VCS,
        if there is one.
        """
        rc, stdout, _ = self.run(
            "-C",
            path,
            "rev-parse",
            "--path-format=absolute",
            "--show-toplevel",
            decode=True,
        )
        out = stdout.strip()
        if rc != 0 or not out:
            return None
        return Path(out)

    def get_current_branch(self, *, repo: Path) -> str | None:
        """
        Retrieve the current branch of the repository.

        If there is no current branch for some reason (detached HEAD state),
        ``None`` is returned.
        """
        rc, stdout, _ = self.run(
            "-C",
            repo,
            "symbolic-ref",
            "--short",
            "--",
            "HEAD",
            check_rc=False,
            decode=True,
        )
        out = stdout.strip()
        if rc != 0 or not out:
            return None
        return out

    def get_changes_compared_to(self, *, repo: Path, branch: str) -> list[Path]:
        """
        Retrieve a list of files and directories that have been added, changed, or removed.

        The returned paths are relative to ``repo``!
        """
        fork_point = self.get_branch_fork_point(repo=repo, branch=branch)
        if fork_point is None:
            raise ValueError(
                f"Cannot determine fork point from current branch/HEAD to {branch}!"
            )

        untracked = self.get_list(
            "ls-files",
            "-z",
            "--others",
            "--exclude-standard",
            repo=repo,
            separator="\0",
        )
        committed = self.get_list(
            "diff",
            "--name-only",
            "--no-renames",
            "-z",
            fork_point,
            "HEAD",
            repo=repo,
            separator="\0",
        )
        staged = self.get_list(
            "diff",
            "--name-only",
            "--no-renames",
            "-z",
            "--cached",
            repo=repo,
            separator="\0",
        )
        unstaged = self.get_list(
            "diff", "--name-only", "--no-renames", "-z", repo=repo, separator="\0"
        )

        all_files = untracked + committed + staged + unstaged
        return [Path(file) for file in sorted(set(all_files))]


__all__ = ("GitProvider",)



# Copyright (c) 2025 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Some basic unit tests.
"""

from __future__ import annotations

from ansible_collections.antsibull.coverage.plugins.modules.test import compute_result


def test_compute_result() -> None:
    """
    Test compute_result().
    """
    result = compute_result()
    assert result == {"msg": "Hello!"}
